import ast
import sys

from pedal.core.commands import system_error
from pedal.core.report import MAIN_REPORT
from pedal.core.location import Location
from pedal.tifa.constants import TOOL_NAME
from pedal.tifa.type_definitions import (UnknownType, RecursedType,
                                         FunctionType, ClassType, InstanceType,
                                         NumType, NoneType, BoolType, TupleType,
                                         ListType, StrType, GeneratorType,
                                         DictType, ModuleType, SetType,
                                         # FileType, DayType, TimeType,
                                         type_from_json, type_to_literal, get_tifa_type,
                                         LiteralNum, LiteralBool,
                                         LiteralNone, LiteralStr,
                                         LiteralTuple)
from pedal.tifa.builtin_definitions import (get_builtin_module, get_builtin_function)
from pedal.tifa.type_operations import (merge_types, are_types_equal,
                                        VALID_UNARYOP_TYPES, VALID_BINOP_TYPES,
                                        ORDERABLE_TYPES, INDEXABLE_TYPES)
from pedal.tifa.identifier import Identifier
from pedal.tifa.state import State
from pedal.tifa.feedbacks import (action_after_return, return_outside_function, write_out_of_scope,
                                  unconnected_blocks, iteration_problem, possible_initialization_problem,
                                  initialization_problem, unused_variable, overwritten_variable,
                                  iterating_over_non_list,
                                  iterating_over_empty_list, incompatible_types, parameter_type_mismatch,
                                  read_out_of_scope, type_changes, unnecessary_second_branch,
                                  recursive_call, not_a_function, incorrect_arity, multiple_return_types,
                                  else_on_loop_body, module_not_found)

__all__ = ['Tifa']


class Tifa(ast.NodeVisitor):
    """
    TIFA Class for traversing an AST and finding common issues.

    Args:
        python_3 (bool): Whether to parse the code in regular PYTHON_3 mode or
                         the modified AST that Skulpt uses.
        report (Report): The report object to store data and feedback in. If
                         left None, defaults to the global MAIN_REPORT.
    """

    def __init__(self, python_3=True, report=MAIN_REPORT):
        self.report = report
        self._initialize_report()

    def _initialize_report(self):
        """
        Initialize a successful report with possible set of issues.
        """
        self.report['tifa'] = {
            'success': True,
            'variables': {},
            'top_level_variables': {},
            'issues': {}
        }

    def locate(self, node: ast.AST = None):
        """
        Return a dictionary representing the current location within the
        AST.

        Returns:
            Location: The line and column of the current or given node.
        """
        if node is None:
            if self.node_chain:
                node = self.node_chain[-1]
            else:
                node = self.final_node
        return Location(node.lineno, col=node.col_offset)

    def process_code(self, code, filename="__main__"):
        """
        Processes the AST of the given source code to generate a report.

        Args:
            code (str): The Python source code
            filename (str): The filename of the source code (defaults to __main__)
        Returns:
            Report: The successful or successful report object
        """
        # Code
        self.source = code.split("\n") if code else []
        filename = filename

        # Attempt parsing - might fail!
        try:
            ast_tree = ast.parse(code, filename)
        except Exception as error:
            self.report['tifa']['success'] = False
            self.report['tifa']['error'] = error
            system_error(TOOL_NAME, "Could not parse code: " + str(error), report=self.report)
            return self.report['tifa']
        try:
            return self.process_ast(ast_tree)
        except Exception as error:
            self.report['tifa']['success'] = False
            self.report['tifa']['error'] = error
            system_error(TOOL_NAME, "Successfully parsed but could not process AST: " + str(error), report=self.report)
            return self.report['tifa']

    def process_ast(self, ast_tree):
        """
        Given an AST, actually performs the type and flow analyses to return a 
        report.

        Args:
            ast_tree (Ast): The AST object
        Returns:
            Report: The final report object created (also available as a field).
        """
        self._reset()
        # Traverse every node
        self.visit(ast_tree)

        # Check afterwards
        self.report['tifa']['variables'] = self.name_map
        self._finish_scope()

        # Collect top level variables
        self._collect_top_level_variables()
        # print(self.report['variables'])

        return self.report['tifa']

    def _issue(self, feedback):
        issues = self.report[TOOL_NAME]['issues']
        if feedback.label not in issues:
            issues[feedback.label] = []
        issues[feedback.label].append(feedback)

    def _collect_top_level_variables(self):
        """
        Walk through the variables and add any at the top level to the
        top_level_variables field of the report.
        """
        top_level_variables = self.report['tifa']['top_level_variables']
        main_path_vars = self.name_map[self.path_chain[0]]
        for full_name in main_path_vars:
            split_name = full_name.split("/")
            if len(split_name) == 2 and split_name[0] == str(self.scope_chain[0]):
                name = split_name[1]
                top_level_variables[name] = main_path_vars[full_name]

    def _reset(self):
        """
        Reinitialize fields for maintaining the system
        """
        # Unique Global IDs
        self.path_id = 0
        self.scope_id = 0
        self.ast_id = 0

        # Human readable names
        self.path_names = ['*Module']
        self.scope_names = ['*Module']
        self.node_chain = []

        # Complete record of all Names
        self.scope_chain = [self.scope_id]
        self.path_chain = [self.path_id]
        self.name_map = {self.path_id: {}}
        self.definition_chain = []
        self.path_parents = {}
        self.final_node = None
        self.class_scopes = {}

    def find_variable_scope(self, name):
        """
        Walk through this scope and all enclosing scopes, finding the relevant
        identifier given by `name`.

        Args:
            name (str): The name of the variable
        Returns:
            Identifier: An Identifier for the variable, which could potentially
                        not exist.
        """
        for scope_level, scope in enumerate(self.scope_chain):
            for path_id in self.path_chain:
                path = self.name_map[path_id]
                full_name = "/".join(map(str, self.scope_chain[scope_level:])) + "/" + name
                if full_name in path:
                    is_root_scope = (scope_level == 0)
                    return Identifier(True, is_root_scope,
                                      full_name, path[full_name])

        return Identifier(False)

    def find_variable_out_of_scope(self, name):
        """
        Walk through every scope and determine if this variable can be found
        elsewhere (which would be an issue).

        Args:
            name (str): The name of the variable
        Returns:
            Identifier: An Identifier for the variable, which could potentially
                        not exist.
        """
        for path in self.name_map.values():
            for full_name in path:
                unscoped_name = full_name.split("/")[-1]
                if name == unscoped_name:
                    return Identifier(True, False, unscoped_name, path[full_name])
        return Identifier(False)

    def find_path_parent(self, path_id, name):
        """

        Args:
            path_id:
            name:

        Returns:

        """
        if name in self.name_map[path_id]:
            return Identifier(True, state=self.name_map[path_id][name])
        else:
            path_parent = self.path_parents.get(path_id)
            if path_parent is None:
                return Identifier(False)
            else:
                return self.find_path_parent(path_parent, name)

    def _finish_scope(self):
        """
        Walk through all the variables present in this scope and ensure that
        they have been read and not overwritten.
        """
        path_id = self.path_chain[0]
        for name in self.name_map[path_id]:
            if Tifa.in_scope(name, self.scope_chain):
                state = self.name_map[path_id][name]
                if state.over == 'yes':
                    position = state.over_position
                    self._issue(overwritten_variable(position, state.name))
                if state.read == 'no':
                    self._issue(unused_variable(state.position, state.name, state.type, report=self.report))

    def visit(self, node):
        """
        Process this node by calling its appropriate visit_*

        Args:
            node (AST): The node to visit
        Returns:
            Type: The type calculated during the visit.
        """
        # Start processing the node
        self.node_chain.append(node)
        self.ast_id += 1

        # Actions after return?
        if len(self.scope_chain) > 1:
            return_state = self.find_variable_scope("*return")
            if return_state.exists and return_state.in_scope:
                if return_state.state.set == "yes":
                    self._issue(action_after_return(self.locate(), report=self.report))

        # No? All good, let's enter the node
        self.final_node = node
        result = ast.NodeVisitor.visit(self, node)

        # Pop the node out of the chain
        self.ast_id -= 1
        self.node_chain.pop()

        # If a node failed to return something, return the UNKNOWN TYPE
        if result is None:
            return UnknownType()
        else:
            return result

    def _visit_nodes(self, nodes):
        """
        Visit all the nodes in the given list.

        Args:
            nodes (list): A list of values, of which any AST nodes will be
                          visited.
        """
        for node in nodes:
            if isinstance(node, ast.AST):
                self.visit(node)

    def walk_targets(self, targets, type, walker):
        """
        Iterate through the targets and call the given function on each one.

        Args:
            targets (list of Ast nodes): A list of potential targets to be
                                         traversed.
            type (Type): The given type to be unraveled and applied to the
                         targets.
            walker (Ast Node, Type -> None): A function that will process
                                             each target and unravel the type.
        """
        for target in targets:
            walker(target, type)

    def _walk_target(self, target, type):
        """
        Recursively apply the type to the target

        Args:
            target (Ast): The current AST node to process
            type (Type): The type to apply to this node
        """
        if isinstance(target, ast.Name):
            self.store_iter_variable(target.id, type, self.locate(target))
            return target.id
        elif isinstance(target, (ast.Tuple, ast.List)):
            result = None
            for i, elt in enumerate(target.elts):
                elt_type = type.iterate(LiteralNum(i))
                potential_name = self._walk_target(elt, elt_type)
                if potential_name is not None and result is None:
                    result = potential_name
            return result

    def visit_AnnAssign(self, node):
        """
        TODO: Implement!
        """
        pass

    def visit_Assign(self, node):
        """
        Simple assignment statement:
        __targets__ = __value__

        Args:
            node (AST): An Assign node
        Returns:
            None
        """
        # Handle value
        value_type = self.visit(node.value)
        # Handle targets
        self._visit_nodes(node.targets)

        self.walk_targets(node.targets, value_type, self.assign_target)

    # TODO: Properly handle assignments with subscripts
    def assign_target(self, target, type):
        """

        Args:
            target:
            type:

        Returns:

        """
        if isinstance(target, ast.Name):
            self.store_variable(target.id, type)
        elif isinstance(target, (ast.Tuple, ast.List)):
            for i, elt in enumerate(target.elts):
                elt_type = type.iterate(LiteralNum(i))
                self.assign_target(elt, elt_type)
        elif isinstance(target, ast.Subscript):
            left_hand_type = self.visit(target.value)
            if isinstance(left_hand_type, ListType):
                # TODO: Handle updating value in list
                pass
            elif isinstance(left_hand_type, DictType):
                if not isinstance(target.slice, ast.Index):
                    # TODO: Can't subscript a dictionary assignment
                    return None
                literal = self.get_literal(target.slice.value)
                if not literal:
                    key_type = self.visit(target.slice.value)
                    left_hand_type.empty = False
                    left_hand_type.keys = [key_type.clone()]
                    left_hand_type.values = [type.clone()]
                elif left_hand_type.literals:
                    original_type = left_hand_type.has_literal(literal)
                    if not original_type:
                        left_hand_type.update_key(literal, type.clone())
                    elif not are_types_equal(original_type, type):
                        # TODO: Fix "Dictionary" to be the name of the variable
                        self._issue(type_changes(self.locate(), 'Dictionary', original_type, type))
        elif isinstance(target, ast.Attribute):
            left_hand_type = self.visit(target.value)
            if isinstance(left_hand_type, InstanceType):
                left_hand_type.add_attr(target.attr, type)
            # TODO: Otherwise we attempted to assign to a non-instance
            # TODO: Handle minor type changes (e.g., appending to an inner list)

    def visit_AugAssign(self, node):
        """

        Args:
            node:

        Returns:

        """
        # Handle value
        right = self.visit(node.value)
        # Handle target
        left = self.visit(node.target)
        # Target is always a Name, Subscript, or Attribute
        name = self.identify_caller(node.target)

        # Handle operation
        self.load_variable(name)
        if isinstance(left, UnknownType) or isinstance(right, UnknownType):
            return UnknownType()
        elif type(node.op) in VALID_BINOP_TYPES:
            op_lookup = VALID_BINOP_TYPES[type(node.op)]
            if type(left) in op_lookup:
                op_lookup = op_lookup[type(left)]
                if type(right) in op_lookup:
                    op_lookup = op_lookup[type(right)]
                    result_type = op_lookup(left, right)
                    self.assign_target(node.target, result_type)
                    return result_type
        self._issue(incompatible_types(self.locate(), node.op, left, right, report=self.report))

    def visit_Attribute(self, node):
        """

        Args:
            node:

        Returns:

        """
        # Handle value
        value_type = self.visit(node.value)
        # Handle ctx
        # TODO: Handling contexts
        # Handle attr
        return value_type.load_attr(node.attr, self, node.value, self.locate())

    def visit_BinOp(self, node):
        """

        Args:
            node:

        Returns:

        """
        # Handle left and right
        left = self.visit(node.left)
        right = self.visit(node.right)

        # Handle operation
        if isinstance(left, UnknownType) or isinstance(right, UnknownType):
            return UnknownType()
        elif type(node.op) in VALID_BINOP_TYPES:
            op_lookup = VALID_BINOP_TYPES[type(node.op)]
            if type(left) in op_lookup:
                op_lookup = op_lookup[type(left)]
                if type(right) in op_lookup:
                    op_lookup = op_lookup[type(right)]
                    return op_lookup(left, right)
        self._issue(incompatible_types(self.locate(), node.op, left, right, report=self.report))
        return UnknownType()

    def visit_Bool(self, node):
        """

        Args:
            node:

        Returns:

        """
        return BoolType()

    def visit_BoolOp(self, node):
        """

        Args:
            node:

        Returns:

        """
        # Handle left and right
        values = []
        for value in node.values:
            values.append(self.visit(value))

        # TODO: Truthiness is not supported! Probably need a Union type
        # TODO: Literals used as truthy value

        # Handle operation
        return BoolType()

    def visit_Call(self, node):
        """

        Args:
            node:

        Returns:

        """
        # Handle func part (Name or Attribute)
        function_type = self.visit(node.func)
        # TODO: Need to grab the actual type in some situations
        callee = self.identify_caller(node)

        # Handle args
        arguments = [self.visit(arg) for arg in node.args] if node.args else []

        # TODO: Handle keywords
        # TODO: Handle starargs
        # TODO: Handle kwargs
        if isinstance(function_type, FunctionType):
            # Test if we have called this definition before
            if function_type.definition not in self.definition_chain:
                self.definition_chain.append(function_type.definition)
                # Function invocation
                result = function_type.definition(self, function_type, callee,
                                                  arguments, self.locate())
                self.definition_chain.pop()
                return result
            else:
                self._issue(recursive_call(self.locate(), callee, report=self.report))
        elif isinstance(function_type, ClassType):
            constructor = function_type.get_constructor().definition
            self.definition_chain.append(constructor)
            result = constructor(self, constructor, callee, arguments, self.locate())
            self.definition_chain.pop()
            if '__init__' in function_type.fields:
                initializer = function_type.fields['__init__']
                if isinstance(initializer, FunctionType):
                    self.definition_chain.append(initializer)
                    initializer.definition(self, initializer, result, [result] + arguments, self.locate())
                    self.definition_chain.pop()
            return result
        else:
            self._issue(not_a_function(self.locate(), callee, report=self.report))
        return UnknownType()

    def visit_ClassDef(self, node):
        """

        Args:
            node:
        """
        class_name = node.name
        new_class_type = ClassType(class_name)
        self.store_variable(class_name, new_class_type)
        # TODO: Define a new scope definition that executes the body
        # TODO: find __init__, execute that
        definitions_scope = self.scope_chain[:]
        class_scope = Tifa.NewScope(self, definitions_scope, class_type=new_class_type)
        with class_scope:
            self.generic_visit(node)

    def visit_Compare(self, node):
        """

        Args:
            node:

        Returns:

        """
        # Handle left and right
        left = self.visit(node.left)
        comparators = [self.visit(compare) for compare in node.comparators]

        # Handle ops
        for op, right in zip(node.ops, comparators):
            if isinstance(op, (ast.Eq, ast.NotEq, ast.Is, ast.IsNot)):
                continue
            elif isinstance(op, (ast.Lt, ast.LtE, ast.GtE, ast.Gt)):
                if are_types_equal(left, right):
                    if isinstance(left, ORDERABLE_TYPES):
                        continue
            elif isinstance(op, (ast.In, ast.NotIn)):
                if isinstance(right, INDEXABLE_TYPES):
                    continue
            self._issue(incompatible_types(self.locate(), op, left, right, report=self.report))
        return BoolType()

    def _visit_collection_loop(self, node):
        # Handle the iteration list
        iter_list = node.iter
        iter_list_name = None
        if isinstance(iter_list, ast.Name):
            iter_list_name = iter_list.id
            if iter_list_name == "___":
                self._issue(unconnected_blocks(self.locate(iter_list), report=self.report))
            state = self.iterate_variable(iter_list_name, self.locate(iter_list))
            iter_type = state.type
        else:
            iter_type = self.visit(iter_list)

        if iter_type.is_empty():
            # TODO: It should check if its ONLY ever iterating over an empty list.
            # For now, only reports if we are NOT in a function
            if len(self.scope_chain) == 1:
                self._issue(iterating_over_empty_list(self.locate(iter_list), iter_list_name))

        if not isinstance(iter_type, INDEXABLE_TYPES):
            self._issue(iterating_over_non_list(self.locate(iter_list), iter_list_name, report=self.report))

        iter_subtype = iter_type.iterate(LiteralNum(0))

        # Handle the iteration variable
        iter_variable_name = self._walk_target(node.target, iter_subtype)

        if iter_variable_name and iter_list_name:
            if iter_variable_name == iter_list_name:
                self._issue(iteration_problem(self.locate(node.target), iter_variable_name))

    def visit_comprehension(self, node):
        """

        Args:
            node:
        """
        self._visit_collection_loop(node)
        # Handle ifs, unless they're blank (None in Skulpt :)
        if node.ifs:
            self.visit_statements(node.ifs)

    def visit_Dict(self, node):
        """
        Three types of dictionaries
        - empty
        - uniform type
        - record
        TODO: Handle records appropriately
        """
        type = DictType()
        if not node.keys:
            type.empty = True
        else:
            type.empty = False
            all_literals = True
            keys, values, literals = [], [], []
            for key, value in zip(node.keys, node.values):
                literal = self.get_literal(key)
                key, value = self.visit(key), self.visit(value)
                values.append(value)
                keys.append(key)
                if literal is not None:
                    literals.append(literal)
                else:
                    all_literals = False

            if all_literals:
                type.literals = literals
                type.values = values
            else:
                type.keys = key
                type.values = value
        return type

    def visit_DictComp(self, node):
        """

        Args:
            node:

        Returns:

        """
        # TODO: Handle comprehension scope
        for generator in node.generators:
            self.visit(generator)
        keys = self.visit(node.key)
        values = self.visit(node.value)
        return DictType(keys=keys, values=values)

    def visit_For(self, node):
        """

        Args:
            node:
        """
        self._visit_collection_loop(node)
        # Handle the bodies
        self.visit_statements(node.body)
        self.visit_statements(node.orelse)

    def visit_FunctionDef(self, node):
        """

        Args:
            node:

        Returns:

        """
        # Name
        function_name = node.name
        position = self.locate()
        definitions_scope = self.scope_chain[:]

        def definition(tifa, call_type, call_name, parameters, call_position):
            """

            Args:
                tifa:
                call_type:
                call_name:
                parameters:
                call_position:

            Returns:

            """
            function_scope = Tifa.NewScope(self, definitions_scope)
            with function_scope:
                # Process arguments
                args = node.args.args
                if len(args) != len(parameters):
                    self._issue(incorrect_arity(self.locate(), function_name, report=self.report))
                # TODO: Handle special types of parameters
                for arg, parameter in zip(args, parameters):
                    name = arg.arg
                    if arg.annotation:
                        self.visit(arg.annotation)
                        annotation = get_tifa_type(arg.annotation, self)
                        # TODO: Use parameter information to "fill in" empty lists
                        if isinstance(parameter, ListType) and isinstance(annotation, ListType):
                            if isinstance(parameter.subtype, UnknownType):
                                parameter.subtype = annotation.subtype
                        # TODO: Check that arg.type and parameter type match!
                        if not are_types_equal(annotation, parameter, True):
                            self._issue(parameter_type_mismatch(self.locate(), name, annotation, parameter))
                    if parameter is not None:
                        parameter = parameter.clone_mutably()
                        self.store_variable(name, parameter, position)
                if len(args) < len(parameters):
                    for undefined_parameter in parameters[len(args):]:
                        self.store_variable(name, UnknownType(), position)
                self.visit_statements(node.body)
                return_state = self.find_variable_scope("*return")
                return_value = NoneType()
                # TODO: Figure out if we are not returning something when we should
                # If the pseudo variable exists, we load it and get its type
                if return_state.exists and return_state.in_scope:
                    return_state = self.load_variable("*return", call_position)
                    return_value = return_state.type
                    if node.returns:
                        # self.visit(node.returns)
                        returns = get_tifa_type(node.returns, self)
                        if not are_types_equal(return_value, returns, True):
                            self._issue(multiple_return_types(return_state.position,
                                                  returns.precise_description(),
                                                  return_value.precise_description(),
                                                  report=self.report))
            return return_value

        function = FunctionType(definition=definition, name=function_name)
        self.store_variable(function_name, function)
        return function

    def visit_GeneratorExp(self, node):
        """

        Args:
            node:

        Returns:

        """
        # TODO: Handle comprehension scope
        for generator in node.generators:
            self.visit(generator)
        return GeneratorType(self.visit(node.elt))

    def visit_If(self, node):
        """

        Args:
            node:
        """
        # Visit the conditional
        self.visit(node.test)

        if len(node.orelse) == 1 and isinstance(node.orelse[0], ast.Pass):
            self._issue(unnecessary_second_branch(self.locate()))
        elif len(node.body) == 1 and isinstance(node.body[0], ast.Pass):
            if node.orelse:
                self._issue(unnecessary_second_branch(self.locate()))

        # Visit the bodies
        this_path_id = self.path_chain[0]
        if_path = Tifa.NewPath(self, this_path_id, "i")
        with if_path:
            for statement in node.body:
                self.visit(statement)
        else_path = Tifa.NewPath(self, this_path_id, "e")
        with else_path:
            for statement in node.orelse:
                self.visit(statement)

        # Combine two paths into one
        # Check for any names that are on the IF path
        self.merge_paths(this_path_id, if_path.id, else_path.id)

    def visit_IfExp(self, node):
        """

        Args:
            node:

        Returns:

        """
        # Visit the conditional
        self.visit(node.test)

        # Visit the body
        body = self.visit(node.body)

        # Visit the orelse
        orelse = self.visit(node.orelse)

        if are_types_equal(body, orelse):
            return body

        # TODO: Union type?
        return UnknownType()

    def visit_Import(self, node):
        """

        Args:
            node:
        """
        # Handle names
        for alias in node.names:
            asname = alias.asname or alias.name
            module_type = self.load_module(alias.name)
            self.store_variable(asname, module_type)

    def visit_ImportFrom(self, node):
        """

        Args:
            node:
        """
        # Handle names
        for alias in node.names:
            if node.module is None:
                asname = alias.asname or alias.name
                module_type = self.load_module(alias.name)
            else:
                module_name = node.module
                asname = alias.asname or alias.name
                module_type = self.load_module(module_name)
            name_type = module_type.load_attr(alias.name, self,
                                              callee_position=self.locate())
            self.store_variable(asname, name_type)

    def visit_Lambda(self, node):
        """

        Args:
            node:

        Returns:

        """
        # Name
        position = self.locate()
        definitions_scope = self.scope_chain[:]

        def definition(tifa, call_type, call_name, parameters, call_position):
            """

            Args:
                tifa:
                call_type:
                call_name:
                parameters:
                call_position:

            Returns:

            """
            function_scope = Tifa.NewScope(self, definitions_scope)
            with function_scope:
                # Process arguments
                args = node.args.args
                if len(args) != len(parameters):
                    self._issue(incorrect_arity(position, "lambda", report=self.report))
                # TODO: Handle special types of parameters
                for arg, parameter in zip(args, parameters):
                    name = arg.arg
                    if parameter is not None:
                        parameter = parameter.clone_mutably()
                        self.store_variable(name, parameter, position)
                if len(args) < len(parameters):
                    for undefined_parameter in parameters[len(args):]:
                        self.store_variable(name, UnknownType(), position)
                return_value = self.visit(node.body)
            return return_value

        return FunctionType(definition=definition)

    def visit_List(self, node):
        """

        Args:
            node:

        Returns:

        """
        type = ListType()
        if node.elts:
            type.empty = False
            # TODO: confirm homogenous subtype
            for elt in node.elts:
                type.subtype = self.visit(elt)
        else:
            type.empty = True
        return type

    def visit_ListComp(self, node):
        """

        Args:
            node:

        Returns:

        """
        # TODO: Handle comprehension scope
        for generator in node.generators:
            self.visit(generator)
        return ListType(self.visit(node.elt))

    def visit_NameConstant(self, node):
        """

        Args:
            node:

        Returns:

        """
        value = node.value
        if isinstance(value, bool):
            return BoolType()
        else:
            return NoneType()

    def visit_Name(self, node):
        """

        Args:
            node:

        Returns:

        """
        name = node.id
        if name == "___":
            self._issue(unconnected_blocks(self.locate()))
        if isinstance(node.ctx, ast.Load):
            if name == "True" or name == "False":
                return BoolType()
            elif name == "None":
                return NoneType()
            else:
                variable = self.find_variable_scope(name)
                builtin = get_builtin_function(name)
                if not variable.exists and builtin:
                    return builtin
                else:
                    state = self.load_variable(name)
                    return state.type
        else:
            variable = self.find_variable_scope(name)
            if variable.exists:
                return variable.state.type
            else:
                return UnknownType()

    def visit_Num(self, node):
        """

        Args:
            node:

        Returns:

        """
        return NumType()

    def visit_Return(self, node):
        """

        Args:
            node:
        """
        if len(self.scope_chain) == 1:
            self._issue(return_outside_function(self.locate(), report=self.report))
        # TODO: Unconditional return inside loop
        if node.value is not None:
            self.return_variable(self.visit(node.value))
        else:
            self.return_variable(NoneType())

    def visit_SetComp(self, node):
        """

        Args:
            node:

        Returns:

        """
        # TODO: Handle comprehension scope
        for generator in node.generators:
            self.visit(generator)
        return SetType(self.visit(node.elt))

    def visit_statements(self, nodes):
        """

        Args:
            nodes:

        Returns:

        """
        # TODO: Check for pass in the middle of a series of statement
        if any(isinstance(node, ast.Pass) for node in nodes):
            pass
        return [self.visit(statement) for statement in nodes]

    def visit_Str(self, node):
        """

        Args:
            node:

        Returns:

        """
        if node.s == "":
            return StrType(True)
        else:
            return StrType(False)

    def visit_Subscript(self, node):
        """

        Args:
            node:

        Returns:

        """
        # Handle value
        value_type = self.visit(node.value)
        # Handle slice
        if isinstance(node.slice, ast.Index):
            literal = self.get_literal(node.slice.value)
            if literal is None:
                dynamic_literal = type_to_literal(self.visit(node.slice.value))
                return value_type.index(dynamic_literal)
            else:
                return value_type.index(literal)
        elif isinstance(node.slice, ast.Slice):
            if node.slice.lower is not None:
                self.visit(node.slice.lower)
            if node.slice.upper is not None:
                self.visit(node.slice.upper)
            if node.slice.step is not None:
                self.visit(node.slice.step)
            return value_type

    def visit_Tuple(self, node):
        """

        Args:
            node:

        Returns:

        """
        type = TupleType()
        if not node.elts:
            type.empty = True
            type.subtypes = []
        else:
            type.empty = False
            # TODO: confirm homogenous subtype
            type.subtypes = [self.visit(elt) for elt in node.elts]
        return type

    def visit_UnaryOp(self, node):
        """

        Args:
            node:

        Returns:

        """
        # Handle operand
        operand = self.visit(node.operand)

        if isinstance(node.op, ast.Not):
            return BoolType()
        elif isinstance(operand, UnknownType):
            return UnknownType()
        elif type(node.op) in VALID_UNARYOP_TYPES:
            op_lookup = VALID_UNARYOP_TYPES[type(node.op)]
            if type(operand) in op_lookup:
                return op_lookup[type(operand)]()
        return UnknownType()

    def visit_While(self, node):
        """

        Args:
            node:
        """
        # Visit conditional
        self.visit(node.test)

        # Visit the bodies
        this_path_id = self.path_id
        # One path is that we never enter the body
        empty_path = Tifa.NewPath(self, this_path_id, "e")
        with empty_path:
            pass
        # Another path is that we loop through the body and check the test again
        body_path = Tifa.NewPath(self, this_path_id, "w")
        with body_path:
            for statement in node.body:
                self.visit(statement)
            # Revisit conditional
            self.visit(node.test)
        # If there's else bodies (WEIRD) then we should check them afterwards
        if node.orelse:
            self._issue(else_on_loop_body(self.locate()))
            for statement in node.orelse:
                self.visit(statement)

        # Combine two paths into one
        # Check for any names that are on the IF path
        self.merge_paths(this_path_id, body_path.id, empty_path.id)

    def visit_With(self, node):
        """

        Args:
            node:
        """
        for item in node.items:
            type_value = self.visit(item.context_expr)
            self.visit(item.optional_vars)
            self._walk_target(item.optional_vars, type_value)
        # Handle the bodies
        self.visit_statements(node.body)

    def _scope_chain_str(self, name=None):
        """
        Convert the current scope chain to a string representation (divided 
        by "/").

        Returns:
            str: String representation of the scope chain.
        """
        if name:
            return "/".join(map(str, self.scope_chain)) + "/" + name
        else:
            return "/".join(map(str, self.scope_chain))

    def identify_caller(self, node):
        """
        Figures out the variable that was used to kick off this call,
        which is almost always the relevant Name to track as being updated.
        If the origin wasn't a Name, nothing will need to be updated so None
        is returned instead.

        TODO: Is this sufficient?

        Args:
            node (AST): An AST node
        Returns:
            str or None: The name of the variable or None if no origin could
                         be found.
        """
        if isinstance(node, ast.Name):
            return node.id
        elif isinstance(node, ast.Call):
            return self.identify_caller(node.func)
        elif isinstance(node, (ast.Attribute, ast.Subscript)):
            return self.identify_caller(node.value)
        return None

    def iterate_variable(self, name, position=None):
        """
        Update the variable by iterating through it - this doesn't do anything
        fancy yet.
        """
        return self.load_variable(name, position)

    def store_iter_variable(self, name, type, position=None):
        """

        Args:
            name:
            type:
            position:

        Returns:

        """
        state = self.store_variable(name, type, position)
        state.read = 'yes'
        return state

    def return_variable(self, return_type):
        """

        Args:
            return_type:

        Returns:

        """
        return self.store_variable("*return", return_type)

    def append_variable(self, name, append_type, position=None):
        """

        Args:
            name:
            append_type:
            position:

        Returns:

        """
        return self.store_variable(name, append_type, position)

    def store_variable(self, name, store_type, position=None):
        """
        Update the variable with the given name to now have the new type.

        Args:
            name (str): The unqualified name of the variable. The variable will
                        be assumed to be in the current scope.
            store_type (Type): The new type of this variable.
            position: The location that this store occurred at
        Returns:
            State: The new state of the variable.
        """
        if position is None:
            position = self.locate()
        full_name = self._scope_chain_str(name)
        current_path = self.path_chain[0]
        variable = self.find_variable_scope(name)
        if not variable.exists:
            # Create a new instance of the variable on the current path
            new_state = State(name, [], store_type, 'store', position,
                              read='no', set='yes', over='no')
            self.name_map[current_path][full_name] = new_state
        else:
            new_state = self.trace_state(variable.state, "store", position)
            if not variable.in_scope:
                self._issue(write_out_of_scope(self.locate(), name, report=self.report))
            # Type change?
            if not are_types_equal(store_type, variable.state.type):
                self._issue(type_changes(position, name, variable.state.type, store_type))
            new_state.type = store_type
            # Overwritten?
            if variable.state.set == 'yes' and variable.state.read == 'no':
                new_state.over_position = position
                new_state.over = 'yes'
            else:
                new_state.set = 'yes'
                new_state.read = 'no'
            self.name_map[current_path][full_name] = new_state
        # If this is a class scope...
        current_scope = self.scope_chain[0]
        if current_scope in self.class_scopes:
            self.class_scopes[current_scope].add_attr(name, new_state.type)
        return new_state

    def load_variable(self, name, position=None):
        """
        Retrieve the variable with the given name.

        Args:
            position:
            name (str): The unqualified name of the variable. If the variable is
                        not found in the current scope or an enclosing sope, all
                        other scopes will be searched to see if it was read out
                        of scope.
        Returns:
            State: The current state of the variable.
        """
        full_name = self._scope_chain_str(name)
        current_path = self.path_chain[0]
        variable = self.find_variable_scope(name)
        if position is None:
            position = self.locate()
        if not variable.exists:
            out_of_scope_var = self.find_variable_out_of_scope(name)
            # Create a new instance of the variable on the current path
            if out_of_scope_var.exists:
                self._issue(read_out_of_scope(self.locate(), name))
            else:
                self._issue(initialization_problem(self.locate(), name))
            new_state = State(name, [], UnknownType(), 'load', position,
                              read='yes', set='no', over='no')
            self.name_map[current_path][full_name] = new_state
        else:
            new_state = self.trace_state(variable.state, "load", position)
            if variable.state.set == 'no':
                self._issue(initialization_problem(self.locate(), name))
            if variable.state.set == 'maybe':
                if name != '*return':
                    self._issue(possible_initialization_problem(self.locate(), name))
            new_state.read = 'yes'
            if not variable.in_scope:
                self.name_map[current_path][variable.scoped_name] = new_state
            else:
                self.name_map[current_path][full_name] = new_state
        return new_state

    def load_module(self, chain):
        """
        Finds the module in the set of available modules.

        Args:
            chain (str): A chain of module imports (e.g., "matplotlib.pyplot")
        Returns:
            ModuleType: The specific module with its members, or an empty
                        module type.
        """
        module_names = chain.split('.')
        potential_module = get_builtin_module(module_names[0])
        if potential_module is not None:
            base_module = potential_module
            for module in module_names:
                if (isinstance(base_module, ModuleType) and
                        module in base_module.submodules):
                    base_module = base_module.submodules[module]
                else:
                    self._issue(module_not_found(self.locate(), chain, False, None, report=self.report))
            return base_module
        else:
            try:
                actual_module = __import__(chain, globals(), {},
                                           ['_tifa_definitions'])
                definitions = actual_module._tifa_definitions()
                return type_from_json(definitions)
            except Exception as e:
                self._issue(module_not_found(self.locate(), chain, True, e, report=self.report))
                return ModuleType()

    def combine_states(self, left, right):
        """

        Args:
            left:
            right:

        Returns:

        """
        state = State(left.name, [left], left.type, 'branch', self.locate(),
                      read=left.read, set=left.set, over=left.over,
                      over_position=left.over_position)
        if right is None:
            state.read = 'no' if left.read == 'no' else 'maybe'
            state.set = 'no' if left.set == 'no' else 'maybe'
            state.over = 'no' if left.over == 'no' else 'maybe'
        else:
            if not are_types_equal(left.type, right.type):
                self._issue(type_changes(self.locate(), left.name, left.type, right.type))
            state.read = Tifa.match_rso(left.read, right.read)
            state.set = Tifa.match_rso(left.set, right.set)
            state.over = Tifa.match_rso(left.over, right.over)
            if left.over == 'no':
                state.over_position = right.over_position
            state.trace.append(right)
        return state

    def merge_paths(self, parent_path_id, left_path_id, right_path_id):
        """
        Combines any variables on the left and right path into the parent
        name space.

        Args:
            parent_path_id (int): The parent path of the left and right branches
            left_path_id (int): One of the two paths
            right_path_id (int): The other of the two paths.
        """
        # Combine two paths into one
        # Check for any names that are on the IF path
        for left_name in self.name_map[left_path_id]:
            left_state = self.name_map[left_path_id][left_name]
            right_identifier = self.find_path_parent(right_path_id, left_name)
            if right_identifier.exists:
                # Was on both IF and ELSE path
                right_state = right_identifier.state
            else:
                # Was only on IF path, potentially on the parent path
                right_state = self.name_map[parent_path_id].get(left_name)
            combined = self.combine_states(left_state, right_state)
            self.name_map[parent_path_id][left_name] = combined
        # Check for names that are on the ELSE path but not the IF path
        for right_name in self.name_map[right_path_id]:
            if right_name not in self.name_map[left_path_id]:
                right_state = self.name_map[right_path_id][right_name]
                # Potentially on the parent path
                parent_state = self.name_map[parent_path_id].get(right_name)
                combined = self.combine_states(right_state, parent_state)
                self.name_map[parent_path_id][right_name] = combined

    def trace_state(self, state, method, position):
        """
        Makes a copy of the given state with the given method type.

        Args:
            position:
            state (State): The state to copy (as in, we trace a copy of it!)
            method (str): The operation being applied to the state.
        Returns:
            State: The new State
        """
        return state.copy(method, position)

    @staticmethod
    def in_scope(full_name, scope_chain):
        """
        Determine if the fully qualified variable name is in the given scope
        chain.

        Args:
            full_name (str): A fully qualified variable name
            scope_chain (list): A representation of a scope chain.
        Returns:
            bool: Whether the variable lives in this scope
        """
        # Get this entity's full scope chain
        name_scopes = full_name.split("/")[:-1]
        # against the reverse scope chain
        checking_scopes = [str(s) for s in scope_chain[::-1]]
        return name_scopes == checking_scopes

    @staticmethod
    def match_rso(left, right):
        """

        Args:
            left:
            right:

        Returns:

        """
        if left == right:
            return left
        else:
            return "maybe"

    def get_literal(self, node):
        """

        Args:
            node:

        Returns:

        """
        if isinstance(node, ast.Num):
            return LiteralNum(node.n)
        elif isinstance(node, ast.Str):
            return LiteralStr(node.s)
        elif isinstance(node, ast.Tuple):
            values = []
            for elt in node.elts:
                subvalue = self.get_literal(elt)
                if subvalue is not None:
                    values.append(subvalue)
                else:
                    return None
            return LiteralTuple(values)
        elif isinstance(node, ast.Name):
            if node.id == "None":
                return LiteralNone()
            elif node.id == "False":
                return LiteralBool(False)
            elif node.id == "True":
                return LiteralBool(True)
        return None

    class NewPath:
        """
        Context manager for entering and leaving execution paths (e.g., if
        statements).)

        Args:
            tifa (Tifa): The tifa instance, so we can modify some of its
                         properties that track variables and paths.
            origin_path (int): The path ID parent to this one.
            name (str): The symbolic name of this path, typically 'i' for an IF
                        body and 'e' for ELSE body.

        Fields:
            id (int): The path ID of this path
        """

        def __init__(self, tifa, origin_path, name):
            self.tifa = tifa
            self.name = name
            self.origin_path = origin_path
            self.id = None

        def __enter__(self):
            self.tifa.path_id += 1
            self.id = self.tifa.path_id
            self.tifa.path_names.append(str(self.id) + self.name)
            self.tifa.path_chain.insert(0, self.id)
            self.tifa.name_map[self.id] = {}
            self.tifa.path_parents[self.id] = self.origin_path

        def __exit__(self, type, value, traceback):
            self.tifa.path_names.pop()
            self.tifa.path_chain.pop(0)

    class NewScope:
        """
        Context manager for entering and leaving scopes (e.g., inside of
        function calls).

        Args:
            tifa (Tifa): The tifa instance, so we can modify some of its
                         properties that track variables and paths.
            definitions_scope_chain (list of int): The scope chain of the
                                                   definition
        """

        def __init__(self, tifa, definitions_scope_chain, class_type=None):
            self.tifa = tifa
            self.definitions_scope_chain = definitions_scope_chain
            self.class_type = class_type

        def __enter__(self):
            # Manage scope
            self.old_scope = self.tifa.scope_chain[:]
            # Move to the definition's scope chain
            self.tifa.scope_chain = self.definitions_scope_chain[:]
            # And then enter its body's new scope
            self.tifa.scope_id += 1
            self.tifa.scope_chain.insert(0, self.tifa.scope_id)
            # Register as class potentially
            if self.class_type is not None:
                self.class_type.scope_id = self.tifa.scope_id
                self.tifa.class_scopes[self.tifa.scope_id] = self.class_type

        def __exit__(self, type, value, traceback):
            # Finish up the scope
            self.tifa._finish_scope()
            # Leave the body
            self.tifa.scope_chain.pop(0)
            # Restore the scope
            self.tifa.scope_chain = self.old_scope



"""
Mocked functions that can be used to prevent malicious or accidental `eval`
behavior.
"""
import re
import types

from pedal.sandbox.exceptions import (SandboxNoMoreInputsException,
                                      SandboxPreventModule)


def _disabled_compile(source, filename, mode, flags=0, dont_inherit=False):
    """
    A version of the built-in `compile` method that fails with a runtime
    error.
    """
    raise RuntimeError("You are not allowed to call 'compile'.")


def _disabled_eval(object, globals=globals(), locals=None):
    """
    A version of the built-in `eval` method that fails with a runtime
    error.
    """
    raise RuntimeError("You are not allowed to call 'eval'.")


# -------------------------------------------------------------


def _disabled_exec(object, globals=globals(), locals=None):
    """
    A version of the built-in `exec` method that fails with a runtime
    error.
    """
    raise RuntimeError("You are not allowed to call 'exec'.")


# -------------------------------------------------------------


def _disabled_globals():
    """
    A version of the built-in `globals` method that fails with a runtime
    error.
    """
    raise RuntimeError("You are not allowed to call 'globals'.")


class FunctionNotAllowed(Exception):
    pass


def disabled_builtin(name):
    """

    Args:
        name:

    Returns:

    """
    def _disabled_version(*args, **kwargs):
        raise FunctionNotAllowed("You are not allowed to call '{}'.".format(name))
    return _disabled_version


_OPEN_FORBIDDEN_NAMES = re.compile(r"(^[./])|(\.py$)")
_OPEN_FORBIDDEN_MODES = re.compile(r"[wa+]")

# TODO: Turn this into a function that lets us more elegantly specify valid and
# invalid filenames/paths


def _restricted_open(name, mode='r', buffering=-1):
    if _OPEN_FORBIDDEN_NAMES.search(name):
        raise RuntimeError("The filename you passed to 'open' is restricted.")
    elif _OPEN_FORBIDDEN_MODES.search(mode):
        raise RuntimeError("You are not allowed to 'open' files for writing.")
    else:
        return _original_builtins['open'](name, mode, buffering)

# TODO: Allow this to be flexible


def _restricted_import(name, globals=None, locals=None, fromlist=(), level=0):
    if name == 'pedal' or name.startswith('pedal.'):
        raise RuntimeError("You cannot import pedal!")
    return _original_builtins['__import__'](name, globals, locals, fromlist, level)


try:
    __builtins__
except NameError:
    _default_builtins = {'globals': globals,
                         'locals': locals,
                         'open': open,
                         'input': input,
                         '__import__': __import__}
else:
    if isinstance(__builtins__, types.ModuleType):
        _default_builtins = __builtins__.__dict__
    else:
        _default_builtins = __builtins__

_original_builtins = {
    'globals': _default_builtins['globals'],
    'locals': _default_builtins['locals'],
    'open': _default_builtins['open'],
    'input': _default_builtins['input'],
    'exec': _default_builtins.get('exec', _disabled_exec),
    'eval': _default_builtins.get('eval', _disabled_eval),
    'compile': _default_builtins.get('compile', _disabled_compile),
    '__import__': _default_builtins['__import__']
}


def make_inputs(input_list, repeat=None):
    """
    Helper function for creating mock user input.

    Params:
        input_list (list of str): The list of inputs to be returned
    Returns:
        function (str=>str): The mock input function that is returned, which
                             will return the next element of input_list each
                             time it is called.
    """
    generator = iter(input_list)

    def mock_input(prompt=''):
        """

        Args:
            prompt:

        Returns:

        """
        print(prompt)
        try:
            return next(generator)
        except StopIteration as SI:
            if repeat is None:
                # TODO: Make this a custom exception
                raise SandboxNoMoreInputsException("User had no more input to give.")
            else:
                return repeat

    return mock_input


_sys_modules = {}


def _override_builtins(namespace, custom_builtins):
    """
    Add the custom builtins to the `namespace` (and the original `__builtins__`)
    suitable for `exec`.
    """
    # Obtain the dictionary of built-in methods, which might not exist in
    # some python versions (e.g., Skulpt)

    # Create a shallow copy of the dictionary of built-in methods. Then,
    # we'll take specific ones that are unsafe and replace them.
    namespace["__builtins__"] = _default_builtins.copy()
    for name, function in custom_builtins.items():
        namespace["__builtins__"][name] = function


def create_module(module_name):
    """

    Args:
        module_name:

    Returns:

    """
    submodule_names = module_name.split(".")
    modules = {}
    root = types.ModuleType(submodule_names[0])
    modules[submodule_names[0]] = root
    reconstructed_path = submodule_names[0]
    for submodule_name in submodule_names[1:]:
        reconstructed_path += "." + submodule_name
        new_submodule = types.ModuleType(reconstructed_path)
        setattr(root, submodule_name, new_submodule)
        modules[reconstructed_path] = new_submodule
    return root, modules


class MockModule:
    def _generate_patches(self):
        return {k: v for k, v in vars(self).items()
                if not k.startswith('_')}

    def _add_to_module(self, module):
        for name, value in self._generate_patches().items():
            setattr(module, name, value)


class BlockedModule(MockModule):
    MODULE_NAME = "this module"

    def _generate_patches(self):
        return {'__getattr__': self.prevent_module}

    def prevent_module(self, **kwargs):
        """

        Args:
            **kwargs:
        """
        raise SandboxPreventModule("You cannot import {module_name} from student code.".format(
            module_name=self.MODULE_NAME
        ))


class MockPedal(BlockedModule):
    MODULE_NAME = "pedal"


class MockTurtle(MockModule):
    """
    Mock Turtle Module that can be used to trace turtle calls.

    Attributes:
        calls (list of dict): The traced list of calls
        # TODO: it'd be awesome to have a way to construct a representation
        #       of the drawing result that we could autograde!
    """
    def __init__(self):
        super().__init__()

    def _reset_turtles(self):
        self.calls = []

    def __repr__(self):
        return repr(self.plots)

    '''
    def _generate_patches(self):
        def dummy(**kwargs):
            pass

        return dict(Canvas=dummy, Pen=dummy, RawPen=dummy, RawTurtle=dummy, Screen=dummy, ScrolledCanvas=dummy,
                    Shape=dummy, TK=dummy, TNavigator=dummy, TPen=dummy, Tbuffer=dummy, Terminator=dummy,
                    Turtle=dummy, TurtleGraphicsError=dummy, TurtleScreen=dummy, TurtleScreenBase=dummy, Vec2D=dummy,
                    addshape=dummy, back=dummy, backward=dummy, begin_fill=dummy, begin_poly=dummy, bgcolor=dummy,
                    bgpic=dummy, bk=dummy, bye=dummy, circle=dummy, clear=dummy, clearscreen=dummy, clearstamp=dummy,
                    clearstamps=dummy, clone=dummy, color=dummy, colormode=dummy, config_dict=dummy, deepcopy=dummy,
                    degrees=dummy, delay=dummy, distance=dummy, done=dummy, dot=dummy, down=dummy, end_fill=dummy,
                    end_poly=dummy, exitonclick=dummy, fd=dummy, fillcolor=dummy, filling=dummy, forward=dummy,
                    get_poly=dummy, get_shapepoly=dummy, getcanvas=dummy, getmethparlist=dummy, getpen=dummy,
                    getscreen=dummy, getshapes=dummy, getturtle=dummy, goto=dummy, heading=dummy, hideturtle=dummy,
                    home=dummy, ht=dummy, inspect=dummy, isdown=dummy, isfile=dummy, isvisible=dummy, join=dummy,
                    left=dummy, listen=dummy, lt=dummy, mainloop=dummy, math=dummy, mode=dummy, numinput=dummy,
                    onclick=dummy, ondrag=dummy, onkey=dummy, onkeypress=dummy, onkeyrelease=dummy, onrelease=dummy,
                    onscreenclick=dummy, ontimer=dummy, pd=dummy, pen=dummy, pencolor=dummy, pendown=dummy,
                    pensize=dummy, penup=dummy, pos=dummy, position=dummy, pu=dummy, radians=dummy,
                    read_docstrings=dummy, readconfig=dummy, register_shape=dummy, reset=dummy, resetscreen=dummy,
                    resizemode=dummy, right=dummy, rt=dummy, screensize=dummy, seth=dummy, setheading=dummy,
                    setpos=dummy, setposition=dummy, settiltangle=dummy, setundobuffer=dummy, setup=dummy,
                    setworldcoordinates=dummy, setx=dummy, sety=dummy, shape=dummy, shapesize=dummy,
                    shapetransform=dummy, shearfactor=dummy, showturtle=dummy, simpledialog=dummy, speed=dummy,
                    split=dummy, st=dummy, stamp=dummy, sys=dummy, textinput=dummy, tilt=dummy, tiltangle=dummy,
                    time=dummy, title=dummy, towards=dummy, tracer=dummy, turtles=dummy, turtlesize=dummy, types=dummy,
                    undo=dummy, undobufferentries=dummy, up=dummy, update=dummy, width=dummy, window_height=dummy,
                    window_width=dummy, write=dummy, write_docstringdict=dummy, xcor=dummy, ycor=dummy)
    '''


class MockPlt(MockModule):
    """
    Mock MatPlotLib library that can be used to capture plot data.

    Attributes:
        plots (list of dict): The internal list of plot dictionaries.
    """

    def __init__(self):
        super().__init__()
        self._reset_plots()

    def show(self, **kwargs):
        """

        Args:
            **kwargs:
        """
        self.plots.append(self.active_plot)
        self._reset_plot()

    def unshown_plots(self):
        """

        Returns:

        """
        return self.active_plot['data']

    def __repr__(self):
        return repr(self.plots)

    def __str__(self):
        return str(self.plots)

    def _reset_plots(self):
        self.plots = []
        self._reset_plot()

    def _reset_plot(self):
        self.active_plot = {'data': [],
                            'xlabel': None, 'ylabel': None,
                            'title': None, 'legend': False}

    def hist(self, data, **kwargs):
        """

        Args:
            data:
            **kwargs:
        """
        label = kwargs.get('label', None)
        self.active_plot['data'].append({'type': 'hist', 'values': data,
                                         'label': label})

    def plot(self, xs, ys=None, **kwargs):
        """

        Args:
            xs:
            ys:
            **kwargs:
        """
        label = kwargs.get('label', None)
        if ys is None:
            self.active_plot['data'].append({'type': 'line',
                                             'x': list(range(len(xs))),
                                             'y': xs, 'label': label})
        else:
            self.active_plot['data'].append({'type': 'line', 'x': xs,
                                             'y': ys, 'label': label})

    def scatter(self, xs, ys, **kwargs):
        """

        Args:
            xs:
            ys:
            **kwargs:
        """
        label = kwargs.get('label', None)
        self.active_plot['data'].append({'type': 'scatter', 'x': xs,
                                         'y': ys, 'label': label})

    def xlabel(self, label, **kwargs):
        """

        Args:
            label:
            **kwargs:
        """
        self.active_plot['xlabel'] = label

    def title(self, label, **kwargs):
        """

        Args:
            label:
            **kwargs:
        """
        self.active_plot['title'] = label

    def suptitle(self, label, **kwargs):
        """

        Args:
            label:
            **kwargs:
        """
        self.title(label, **kwargs)

    def ylabel(self, label, **kwargs):
        """

        Args:
            label:
            **kwargs:
        """
        self.active_plot['ylabel'] = label

    def legend(self, **kwargs):
        """

        Args:
            **kwargs:
        """
        self.active_plot['legend'] = True

    def _generate_patches(self):
        def dummy(**kwargs):
            """

            Args:
                **kwargs:
            """
            pass

        return dict(hist=self.hist, plot=self.plot,
                    scatter=self.scatter, show=self.show,
                    xlabel=self.xlabel, ylabel=self.ylabel,
                    title=self.title, legend=self.legend,
                    xticks=dummy, yticks=dummy,
                    autoscale=dummy, axhline=dummy,
                    axhspan=dummy, axvline=dummy,
                    axvspan=dummy, clf=dummy,
                    cla=dummy, close=dummy,
                    figlegend=dummy, figimage=dummy,
                    suptitle=self.suptitle, text=dummy,
                    tick_params=dummy, ticklabel_format=dummy,
                    tight_layout=dummy, xkcd=dummy,
                    xlim=dummy, ylim=dummy,
                    xscale=dummy, yscale=dummy)



"""
A tool for running Pedal from the Command Line

./pedal command line interface
    simple example (single ICS, single student file)
    simple sandbox for just quickly running students code with no frills or extra behavior
        Maybe this is the "default" ICS?
    unit test a given curriculum (many ICS, many student files per)
    regrade some assignments, spit out subject/assignment/context (one ICS, many many student files)
    get summary stats from ProgSnap file (many ICS, many submissions
    Spit out the justifications alongside the potential feedback`
    Run the students code in a sandbox and spit out the runtime results

ProgSnap file with submissions AND assignments (+other info)
ProgSnap file with only submissions, ICS is separate
Single ICS file
    Single student file
    Archive Format
        Regular directory
        ZIP
        ProgSnap Folder
        ProgSnap SQL
    Structure
        Each file is a python file representing their submission:
            .+\.py
            Could have accompanying Markdown file
        Each folder is a submission
            .+/.+\.py
Folder of ICS files
    Potentially, submissions are within an accompanying "submissions" folder
    Regex for matching ICS files?



argparse options
    -i <instructor_control_script | python file or directory of scripts>
    -s <student_submission | single python file, directory of files, progsnap file>
    -o <output_file>
    -f <output format: Feedback, Grade, Debug mode, Unit Tests, research stats, Sandbox>

For each ICS we want to run
    For each student submission
        safely run it
    Potentially combine results
"""

import argparse
import os
import sys
from contextlib import redirect_stdout

from io import StringIO
from textwrap import indent

import unittest
from unittest.mock import patch

from pedal import clear_report
from pedal.core.report import MAIN_REPORT
from pedal.cait import parse_program
from pedal.command_line.mocks import MockOpen


def normalize_path(first: str, second: str):
    """
    If the `first` path is relative, creates an absolute path relative to the `second` folder.

    Args:
        first: The path to normalize
        second: The path to normalize in regards to.

    Returns:
        str: A new path.
    """
    if os.path.isabs(first):
        return first
    else:
        return os.path.join(os.path.dirname(second), first)


def find_seed(python_code):
    """
    Attempts to find a top-level variable named "__STUDENT_SEED__" inside the given python code.
    Args:
        python_code (str): A string of valid Python code.
    Returns:
        str or int or List[int]: The seed found in the file, or 0.
    """
    try:
        ast = parse_program(python_code)
        for assign in ast.find_all("Assign"):
            if assign.targets[0].ast_name != "Name":
                continue
            if assign.targets[0].id == "__STUDENT_SEED__":
                if assign.value.ast_name == "Str":
                    return assign.value.s
                elif assign.value.ast_name == "Num":
                    return assign.value.n
                elif assign.value.ast_name == "List":
                    return [e.n for e in assign.value.elts]
    except SyntaxError:
        return 0
    return 0


def substitute_args(args, student_path, seed):
    """
    Updates the given arguments by replacing $template values with the given ones.

    Args:
        arg (List[str]):
        student_path (str):
        seed (Any):

    Returns:

    """
    for arg in args.split(","):
        if arg == "$_STUDENT_MAIN":
            yield student_path
        elif arg == "$_STUDENT_NAME":
            yield seed
        yield arg


def run_ics_on_submission(ics_filename: str, ics_contents: str, ics_args: str,
                          submission_filename: str, submission_contents: str,
                          output_format: str,
                          student: dict, course: dict, assignment: dict):
    """
    Runs the given instructor control script on the given submission, with the accompany contextualizations.

    Args:
        ics_filename (str):
        ics_contents (str):
        ics_args (str): Additional arguments for the ICS script.
        submission_filename (str):
        submission_contents (str):
        output_format (str): The format to use when running
        student (dict):
        course (dict):
        assignment (dict):

    Returns:

    """
    seed = find_seed(submission_contents)
    # TODO: figure out args
    #ics_args = list(substitute_args(ics_args, submission_filename, seed))
    ics_args = [ics_filename, {submission_filename: submission_contents},
                submission_filename, submission_contents,
                student, assignment, course, None]
    captured_output = StringIO()
    global_data = {}
    with redirect_stdout(captured_output):
        with patch('builtins.open', MockOpen(read_data=submission_contents),
                   create=True):
            with patch.object(sys, 'argv', ics_args):
                clear_report()
                grader_exec = compile(ics_contents, ics_filename, 'exec')
                exec(grader_exec, global_data)
    actual_output = captured_output.getvalue()
    return {"output": actual_output, "data": global_data, "report": MAIN_REPORT}


class Result:
    def __init__(self):
        self.results = []

    def add_case(self, submission_file_name, contents, result):
        self.results.append((submission_file_name, contents, result))
        if format == FORMAT_OPTIONS.FEEDBACK:
            print(result['filename'])
            print(indent(result['report'].result.for_console(), " " * 4))
        elif format == FORMAT_OPTIONS.GRADE:
            print(result['report'].result.score)
        elif format == FORMAT_OPTIONS.DEBUG:
            pass
        elif format == FORMAT_OPTIONS.STATS:
            assert "TODO: Aggregation of feedback into stats file is not yet supported"


class VerifyResults(Result):
    def __init__(self):
        super().__init__()

        class TestReferenceSolutions(unittest.TestCase):
            maxDiff = None

        self.TestReferenceSolutions = TestReferenceSolutions
        self.tests = []

    def add_case(self, submission_file_name, contents, result):
        with open(submission_file_name + ".txt", "r") as output_file:
            output_contents = output_file.read()

        if result and result[-1] == "\n":
            result = result[:-1]

        def new_test(self):
            self.assertEqual(result, output_contents)
        test_name = os.path.basename(submission_file_name)
        setattr(self.TestReferenceSolutions, "test_"+test_name, new_test)
        self.tests.append("test_"+test_name)
        # assert "TODO: Unit test style verification not yet supported"

    def run_cases(self):
        suite = unittest.defaultTestLoader.loadTestsFromTestCase(self.TestReferenceSolutions)
        runner = unittest.TextTestRunner()
        runner.run(suite)


class FORMAT_OPTIONS:
    """
    The possible valid output options
    """
    FEEDBACK = 'feedback'
    GRADE = 'grade'
    DEBUG = 'debug'
    VERIFY = 'verify'
    STATS = 'stats'
    SANDBOX = 'sandbox'
    ALL_OPTIONS = [FEEDBACK, GRADE, DEBUG, VERIFY, STATS, SANDBOX]
    CONSTRUCTORS = {
        VERIFY: VerifyResults
    }


def main(args=None):
    """
    Actually runs Pedal from the command line.

    Args:
        args:

    Returns:

    """
    if args is None:
        args = parse_args()
    result = FORMAT_OPTIONS.CONSTRUCTORS[args.format]()
    if os.path.isfile(args.script):
        script_file_name, script_file_extension = os.path.splitext(args.script)
        if script_file_extension in ('.py',):
            submissions = normalize_path(args.submissions, args.script)
            if os.path.isfile(submissions):
                submission_paths = [submissions]
            else:
                submission_paths = [normalize_path(sub, submissions) for sub in os.listdir(submissions)]
            for submission_path in submission_paths:
                submission_file_name, submission_file_extension = os.path.splitext(submission_path)
                if submission_file_extension != ".py":
                    continue
                # TODO: add in student/course/assignment info
                with open(args.script, 'r') as scripts_file:
                    scripts_contents = scripts_file.read()
                with open(submission_path, 'r') as submission_file:
                    submission_contents = submission_file.read()
                print(submission_file_name)
                single_result = run_ics_on_submission(args.script, scripts_contents, args.args,
                                               "answer.py", submission_contents,
                                               args.format,
                                               {'filename': submission_path}, {}, {})
                single_result['filename'] = submission_path
                result.add_case(submission_file_name, submission_contents, single_result['output'])
            result.run_cases()
        elif script_file_extension in ('.db',):
            assert "TODO: ProgSnap DB files not yet supported"
        elif script_file_extension in ('.zip', ):
            assert "TODO: Zip files not yet supported"
    else:
        assert "TODO: Folders of grading scripts not yet supported"


def parse_args():
    parser = argparse.ArgumentParser(description='Run instructor control script on student submissions.')
    parser.add_argument('format', help='The format for the results',
                        choices=FORMAT_OPTIONS.ALL_OPTIONS)
    parser.add_argument('script', help='The path to the instructor control script, or multiple scripts.')
    parser.add_argument('submissions', help='The path to the student submissions. Defaults to a folder named '
                                            'submissions adjacent to the instructor control script.',
                        default='submissions')
    parser.add_argument('output', help='The output path for the result',
                        nargs='?', default=None)
    parser.add_argument('--args', '-a',
                        help='Pass in arguments that the grading script will use. '
                             'Variable substitutions include "$_STUDENT_MAIN".',
                        default='$_STUDENT_MAIN,$_STUDENT_NAME')
    parser.add_argument('--include_submissions', help='An optional REGEX filter to only include certain submissions')
    parser.add_argument('--exclude_submissions', help='An optional REGEX filter to remove certain submissions')
    parser.add_argument('--include_scripts', help='An optional REGEX filter to only include certain scripts')
    parser.add_argument('--exclude_scripts', help='An optional REGEX filter to remove certain scripts')
    parser.add_argument('--parallel_scripts', help="Which style to use for running scripts in parallel.",
                        choices=["threads", "processes", "none"])
    args = parser.parse_args()
    return args

if __name__ == '__main__':
    main()

# If scripts is a single python file
# If scripts is a folder
# If scripts is a zip file
# If scripts is a ProgSnap SQL file



from pedal import feedback
from pedal.core.feedback import Feedback, AtomicFeedbackFunction
from pedal.core.commands import give_partial
from pedal.core.report import MAIN_REPORT
from pedal.sandbox.result import SandboxResult
from pedal.sandbox.sandbox import DataSandbox
from pedal.assertions.setup import _setup_assertions, AssertionException, TOOL_NAME_ASSERTIONS
from pedal.utilities.comparisons import equality_test, output_test
from pedal.utilities.types import humanize_types


# TODO: Allow bundling of assertions to make a table


def iterable(obj):
    """

    Args:
        obj:

    Returns:

    """
    return hasattr(obj, '__iter__') or hasattr(obj, '__getitem__')


DELTA = .001
_MAX_LENGTH = 80


def _escape_curly_braces(result):
    return result.replace("{", "{{").replace("}", "}}")


def safe_repr(obj, short=False):
    """

    Args:
        obj:
        short:

    Returns:

    """
    try:
        result = repr(obj)
    except Exception:
        result = object.__repr__(obj)
    if short and len(result) >= _MAX_LENGTH:
        result = result[:_MAX_LENGTH] + ' [truncated]...'
    result = result
    return result


def _fail(code_message, actual_message, expected_message,
          show_expected_value, modify_right, *values):
    normal_values = []
    sandboxed_values = []
    sandboxed_results = []
    if modify_right and values:
        values = values[:-1] + (modify_right(values[-1]),)
    for value in values:
        if is_sandbox_result(value):
            sandboxed_results.append(value)
            value = value._actual_value
            sandboxed_values.append(safe_repr(value))
        else:
            normal_values.append(safe_repr(value))
    if sandboxed_results:
        code_message = _build_context(sandboxed_results, actual_message,
                                      expected_message, show_expected_value)
    return AssertionException(code_message.format(*(sandboxed_values + normal_values)))


def _build_result_from_target(target, index, quantity):
    if target == "_":
        if quantity == 1:
            return "the result"
        elif index == 0:
            return "the first result"
        else:
            return "the second result"
    return "<code>" + target + "</code>"


def _build_context(sandboxed_results, actual_message, expected_message,
                   show_expected_value):
    context = []
    calls = []
    inputs = []
    outputs = []
    targets = []
    for result in sandboxed_results:
        # Look up info
        call_id = result._actual_call_id
        sandbox = result._actual_sandbox
        outputs.extend(sandbox.output_contexts[call_id])
        calls.extend(sandbox.call_contexts[call_id])
        inputs.extend(sandbox.input_contexts[call_id])
        targets.append(sandbox.target_contexts[call_id])
    # Actual rendering of text
    if calls:
        calls = [_escape_curly_braces(str(call)) for call in calls]
        context.append("I ran:\n<pre>" + "\n".join(calls) + "</pre>")
    if inputs:
        inputs = [_escape_curly_braces(str(inp)) for inp in inputs]
        context.append("I entered as input:\n<pre>" + "\n".join(inputs) + "</pre>")
    actual_message += ":\n<pre>{}</pre>"
    for i, target in enumerate(targets):
        named_target = _build_result_from_target(target, i, len(targets))
        if target == '_':
            context.append(named_target.capitalize() + " " + actual_message)
        else:
            context.append("The value of " + named_target + " " + actual_message)
    expected_context = "But I expected "
    if len(targets) == 2:
        expected_context += _build_result_from_target(targets[0], 0, 2)
        expected_context += " " + expected_message + " "
        expected_context += _build_result_from_target(targets[1], 1, 2)
    else:
        expected_context += _build_result_from_target(targets[0], 0, 1)
        expected_context += " " + expected_message
        if show_expected_value:
            expected_context += ":\n<pre>{}</pre>"
    context.append(expected_context)
    return "\n".join(context)


def is_sandbox_result(value):
    """

    Args:
        value:

    Returns:

    """
    if hasattr(value, "__actual_class__"):
        if value.__actual_class__ == SandboxResult:
            return True
    return False


@AtomicFeedbackFunction(title="Instructor Test",
                        message_template=("Student code failed instructor test.\n"
                                          "{context}{failure}{message}"))
def _basic_assertion(label, justification, left, right, operator, code_comparison_message,
                     hc_message, hc_message_past, message, contextualize, report=MAIN_REPORT,
                     show_expected_value=True, modify_right=None, partial_score=0, muted=False):
    _setup_assertions(report)
    context = ""
    if message:
        message = "\n" + message
    else:
        message = ""
    # TODO: Handle right-side sandbox result
    # if is_sandbox_result(right):
    #    right = right._actual_value
    if isinstance(left, Exception):
        return False
    if isinstance(right, Exception):
        return False
    if not operator(left, right):
        failure = _fail(code_comparison_message, hc_message, hc_message_past,
                        show_expected_value, modify_right, left, right)
        report[TOOL_NAME_ASSERTIONS]['collected'].append(failure)
        fields = {'context': context, 'failure': str(failure), 'message': message}
        feedback(label, tool=TOOL_NAME_ASSERTIONS, category=Feedback.CATEGORIES.INSTRUCTOR,
                 justification=justification, title=_basic_assertion.title,
                 message=_basic_assertion.message_template.format(**fields), fields=fields,
                 score=partial_score, correct=False, muted=muted, report=report)
        report[TOOL_NAME_ASSERTIONS]['failures'] += 1
        if report[TOOL_NAME_ASSERTIONS]['exceptions']:
            raise failure
        else:
            return False
    return True


def assertEqual(left, right, score=None, message=None, report=MAIN_REPORT,
                contextualize=True, exact=False, compare_lengths=None, muted=False,
                label="assert_equal"):
    """

    Args:
        label:
        left:
        right:
        score:
        message:
        report:
        contextualize:
        exact:
        compare_lengths:
        muted:

    Returns:

    """
    if compare_lengths is None:
        compare_lengths = (iterable(left) and isinstance(right, (int, float)))
    return _basic_assertion(label, "Left and right were not equal", left, right,
                            lambda l, r:
                            equality_test(len(l), r, exact, DELTA) if
                            compare_lengths else
                            equality_test(l, r, exact, DELTA),
                            "len({}) != {}" if compare_lengths else "{} != {}",
                            "was",
                            "to have its length equal to"
                            if compare_lengths else "to be equal to",
                            message, contextualize, report=report, partial_score=score, muted=muted)


def assertNotEqual(left, right, score=None, message=None, report=MAIN_REPORT,
                   contextualize=True, exact=False, label="assert_not_equal", muted=False):
    """

    Args:
        left:
        right:
        score:
        message:
        report:
        contextualize:
        exact:

    Returns:

    """
    return _basic_assertion(label, "Left and right were equal", left, right,
                            lambda l, r: not equality_test(l, r, exact, DELTA),
                            "{} == {}",
                            "was",
                            "to not be equal to",
                            message, contextualize, report=report, partial_score=score, muted=muted)


def assertTrue(something, score=None, message=None, report=MAIN_REPORT,
               contextualize=True, label="assert_true", muted=False):
    """

    Args:
        something:
        score:
        message:
        report:
        contextualize:

    Returns:

    """
    return _basic_assertion(label, "Value was not True", something, True,
                            lambda l, r: bool(l),
                            "{} is true",
                            "was false",
                            "to be true",
                            message, contextualize, report=report,
                            show_expected_value=False, partial_score=score, muted=muted)


def assertFalse(something, score=None, message=None, report=MAIN_REPORT,
                contextualize=True, label="assert_false", muted=False):
    """

    Args:
        something:
        score:
        message:
        report:
        contextualize:

    Returns:

    """
    return _basic_assertion(label, "Value was not False", something, False,
                            lambda l, r: not bool(l),
                            "{} is false",
                            "was true",
                            "to be false",
                            message, contextualize, report=report, partial_score=score, muted=muted,
                            show_expected_value=False)


def assertIs(left, right, score=None, message=None, report=MAIN_REPORT, contextualize=True,
             label="assert_is", muted=False):
    """

    Args:
        left:
        right:
        score:
        message:
        report:
        contextualize:
    """
    raise NotImplementedError()


def assertIsNot(left, right, score=None, message=None, report=MAIN_REPORT, contextualize=True,
                label="assert_is_not", muted=False):
    """

    Args:
        left:
        right:
        score:
        message:
    """
    raise NotImplementedError()


def _actually_is_none(l, r):
    if is_sandbox_result(l):
        return l._actual_value is None
    return l is None


def assertIsNone(something, score=None, message=None, report=MAIN_REPORT,
                 contextualize=True, label="assert_is_none", muted=False):
    """

    Args:
        something:
        score:
        message:
        report:
        contextualize:

    Returns:

    """
    return _basic_assertion(label, "Value is not None", something, None,
                            _actually_is_none,
                            "{} is none",
                            "was",
                            "to be none",
                            message, contextualize, report=report, partial_score=score, muted=muted,
                            show_expected_value=False)


def _actually_is_not_none(l, r):
    if is_sandbox_result(l):
        return l._actual_value is not None
    return l is not None


def assertIsNotNone(something, score=None, message=None, report=MAIN_REPORT,
                    contextualize=True, label="assert_is_not_none", muted=False):
    """

    Args:
        something:
        score:
        message:
        report:
        contextualize:

    Returns:

    """
    return _basic_assertion(label, "Value is None", something, None,
                            _actually_is_not_none,
                            "{} is not none",
                            "was",
                            "to not be none",
                            message, contextualize, report=report, partial_score=score, muted=muted,
                            show_expected_value=False)


def assertIn(needle, haystack, score=None, message=None, report=MAIN_REPORT,
             contextualize=True, label="assert_in", muted=False):
    """

    Args:
        needle:
        haystack:
        score:
        message:
        report:
        contextualize:

    Returns:

    """
    expected_message = "to be in"
    if not is_sandbox_result(needle) and is_sandbox_result(haystack):
        expected_message = "to contain"
    return _basic_assertion(label, "Needle not in haystack", needle, haystack,
                            lambda n, h: n in h,
                            "{} not in {}",
                            "was",
                            expected_message,
                            message, contextualize, report=report, partial_score=score, muted=muted)


def assertNotIn(needle, haystack, score=None, message=None, report=MAIN_REPORT,
                contextualize=True, label="assert_not_in", muted=False):
    """

    Args:
        needle:
        haystack:
        score:
        message:
        report:
        contextualize:

    Returns:

    """
    expected_message = "to not be in"
    if not is_sandbox_result(needle) and is_sandbox_result(haystack):
        expected_message = "to not contain"
    return _basic_assertion(label, "Needle in haystack", needle, haystack,
                            lambda n, h: n not in h,
                            "{} in {}",
                            "was",
                            expected_message,
                            message, contextualize, report=report, partial_score=score, muted=muted)


def assertIsInstance(value, types, score=None, message=None, report=MAIN_REPORT,
                     contextualize=True, label="assert_is_instance", muted=False):
    """

    Args:
        value:
        types:
        score:
        message:
        report:
        contextualize:

    Returns:

    """
    return _basic_assertion(label, "Value is not of type", value, types,
                            lambda v, t: isinstance(v, t),
                            "isinstance({}, {})",
                            "was",
                            "to be of type",
                            message, contextualize, report=report, partial_score=score, muted=muted,
                            modify_right=humanize_types)


def assertNotIsInstance(value, types, score=None, message=None, report=MAIN_REPORT,
                        contextualize=True, label="assert_not_is_instance", muted=False):
    """

    Args:
        value:
        types:
    """
    return _basic_assertion(label, "Value is of the type", value, types,
                            lambda v, t: not isinstance(v, t),
                            "not isinstance({}, {})",
                            "was",
                            "to not be of type",
                            message, contextualize, report=report, partial_score=score, muted=muted,
                            modify_right=humanize_types)


def assertRaises(exception):
    """

    Args:
        exception:
    """
    raise NotImplementedError()


def assertRaisesRegexp(exception):
    """

    Args:
        exception:
    """
    raise NotImplementedError()


def assertAlmostEqual(left, right):
    """

    Args:
        left:
        right:
    """
    raise NotImplementedError()


def assertNotAlmostEqual(left, right):
    """

    Args:
        left:
        right:
    """
    raise NotImplementedError()


def assertGreater(left, right, score=None, message=None, report=MAIN_REPORT, muted=False,
                  contextualize=True, compare_lengths=None, label="assert_greater"):
    """

    Args:
        left:
        right:
        score:
        message:
        report:
        contextualize:
        compare_lengths:

    Returns:

    """
    if compare_lengths is None:
        compare_lengths = (iterable(left) and isinstance(right, (int, float)))
    return _basic_assertion(label, "Left is not greater than right", left, right,
                            lambda l, r:
                            len(l) > r if
                            compare_lengths else
                            l > r,
                            "len({}) <= {}" if compare_lengths else "{} <= {}",
                            "was",
                            "to have its length greater than"
                            if compare_lengths else
                            "to be greater than",
                            message, contextualize, report=report, partial_score=score, muted=muted)


def assertGreaterEqual(left, right, score=None, message=None, report=MAIN_REPORT, muted=False,
                       contextualize=True, compare_lengths=None, label="assert_greater_equal"):
    """

    Args:
        left:
        right:
        score:
        message:
        report:
        contextualize:
        compare_lengths:

    Returns:

    """
    if compare_lengths is None:
        compare_lengths = (iterable(left) and isinstance(right, (int, float)))
    return _basic_assertion(label, "Left is not greater or equal than right", left, right,
                            lambda l, r:
                            len(l) >= r if
                            compare_lengths else
                            l >= r,
                            "len({}) < {}" if compare_lengths else "{} < {}",
                            "was",
                            "to have its length greater than or equal to" if compare_lengths else
                            "to be greater than or equal to",
                            message, contextualize, report=report, partial_score=score, muted=muted)


def assertLess(left, right, score=None, message=None, report=MAIN_REPORT, muted=False,
               contextualize=True, compare_lengths=None, label="assert_less"):
    """

    Args:
        left:
        right:
        score:
        message:
        report:
        contextualize:
        compare_lengths:

    Returns:

    """
    if compare_lengths is None:
        compare_lengths = (iterable(left) and isinstance(right, (int, float)))
    return _basic_assertion(label, "Left is not less than right", left, right,
                            lambda l, r:
                            len(l) < r if
                            compare_lengths else
                            l < r,
                            "len({}) >= {}" if compare_lengths else "{} >= {}",
                            "was",
                            "to have its length less than"
                            if compare_lengths else
                            "to be less than",
                            message, contextualize, report=report, partial_score=score, muted=muted)


def assertLessEqual(left, right, score=None, message=None, report=MAIN_REPORT, muted=False,
                    contextualize=True, compare_lengths=None, label="assert_less_equal"):
    """

    Args:
        left:
        right:
        score:
        message:
        report:
        contextualize:
        compare_lengths:

    Returns:

    """
    if compare_lengths is None:
        compare_lengths = (iterable(left) and isinstance(right, (int, float)))
    return _basic_assertion(label, "Left is not less than or equal to right", left, right,
                            lambda l, r:
                            len(l) <= r if
                            compare_lengths else
                            l <= r,
                            "len({}) > {}" if compare_lengths else "{} > {}",
                            "was",
                            "to have its length less than or equal to" if compare_lengths else
                            "to be less than or equal to",
                            message, contextualize, report=report, partial_score=score, muted=muted)


def assertRegexpMatches(text, pattern):
    """

    Args:
        text:
        pattern:
    """
    raise NotImplementedError()


def assertNotRegexpMatches(text, pattern):
    """

    Args:
        text:
        pattern:
    """
    raise NotImplementedError()


def assertItemsEqual(left, right):
    """

    Args:
        left:
        right:
    """
    raise NotImplementedError()


def assertDictContainsSubset(left, right):
    """

    Args:
        left:
        right:
    """
    raise NotImplementedError()


def assertMultiLineEqual(left, right):
    """

    Args:
        left:
        right:
    """
    raise NotImplementedError()


def assertSequenceEqual(left, right):
    """

    Args:
        left:
        right:
    """
    raise NotImplementedError()


# TODO: assertPrintIncludes

# Speciality Asserts
def assertPrints(result, expected_output, args=None, returns=None,
                 score=None, message=None, report=MAIN_REPORT, muted=False,
                 contextualize=True, exact=False, label="assert_prints"):
    """

    Args:
        result:
        expected_output:
        args:
        returns:
        score:
        message:
        report:
        contextualize:
        exact:

    Returns:

    """
    if not isinstance(result, SandboxResult):
        return False
        raise TypeError("You must pass in a SandboxResult (e.g., using `call`) to assertPrints")
    _setup_assertions(report)
    call_id = result._actual_call_id
    sandbox = result._actual_sandbox
    calls = sandbox.call_contexts[call_id]
    inputs = sandbox.input_contexts[call_id]
    actual_output = sandbox.output_contexts[call_id]
    if not output_test(actual_output, expected_output, exact):
        context = []
        if calls:
            context.append("I ran:\n<pre>" +
                           "\n".join(map(str, calls)) +
                           "</pre>")
        if inputs:
            context.append("I entered as input:\n<pre>" +
                           "\n".join(map(str, inputs)) +
                           "</pre>")
        if actual_output:
            context.append("The function printed:\n<pre>" +
                           "\n".join(map(str, actual_output)) +
                           "</pre>")
        else:
            context.append("The function printed nothing.")
        context.append("But I expected the output:\n<pre>" + "\n".join(map(str, expected_output)) + "</pre>")
        failure = AssertionException("\n".join(context))
        report['assertions']['collected'].append(failure)
        # TODO: Fix this!
        report.attach('Instructor Test', category='student', tool='Assertions',
                      mistake={'message': "Student code failed instructor test.<br>\n" +
                                          str(failure)})
        report['assertions']['failures'] += 1
        if report['assertions']['exceptions']:
            raise failure
        else:
            return False
    give_partial(score)
    return True


def assertHasFunction(obj, function, args=None, returns=None,
                      score=None, message=None, report=MAIN_REPORT, muted=False,
                      contextualize=True, exact=False, label="assert_has_function"):
    """

    Args:
        obj:
        function:
        args:
        returns:
        score:
        message:
        report:
        contextualize:
        exact:

    Returns:

    """
    # If object is a sandbox, will check the .data[variable] attribute
    # Otherwise, check it directly
    if isinstance(obj, DataSandbox):
        comparison = lambda o, f: f in o.data
    else:
        def comparison(o, f):
            """

            Args:
                o:
                f:

            Returns:

            """
            try:
                return f in o
            except:
                return hasattr(o, f)
    if not _basic_assertion(label, "Student object does not have function", obj, function,
                            comparison,
                            "Could not find function {}{}",
                            "was",
                            "to have the function",
                            message, contextualize=contextualize, report=report, partial_score=score, muted=muted):
        return False
    if isinstance(obj, DataSandbox):
        student_function = obj.data[function]
    else:
        try:
            student_function = obj[function]
        except:
            student_function = getattr(obj, function)
    if _basic_assertion(label, "Value is not a function", student_function, function,
                        lambda l, r: callable(l),
                        "The value {} is in the variable {}, and that value is not a callable function.",
                        "was callable",
                        "to be callable",
                        message, contextualize=contextualize, report=report, partial_score=score, muted=muted,
                        show_expected_value=False):
        give_partial(score)
        return True
    return False


def assertHas(obj, variable, types=None, value=None, score=None, muted=False,
              message=None, report=MAIN_REPORT, contextualize=True, label="assert_has"):
    """

    Args:
        obj:
        variable:
        types:
        value:
        score:
        message:
        report:
        contextualize:

    Returns:

    """
    # If object is a sandbox, will check the .data[variable] attribute
    # Otherwise, check it directly
    if isinstance(obj, DataSandbox):
        comparison = lambda o, v: v in o.data
    else:
        comparison = lambda o, v: v in hasattr(o, v)
    if not _basic_assertion(label, "Object does not have variable", obj, variable,
                            comparison,
                            "Could not find variable {}{}",
                            "was",
                            "to have the variable",
                            message, contextualize, report=report, partial_score=score, muted=muted):
        return False
    if isinstance(obj, DataSandbox):
        student_variable = obj.data[variable]
    else:
        student_variable = getattr(obj, variable)
    if types is not None:
        if not _basic_assertion(label, "Object's value does not have the right type", student_variable, types,
                                lambda v, t: isinstance(v, t),
                                "isinstance({}, {})",
                                "was",
                                "to be of type",
                                message, contextualize, report=report, partial_score=score, muted=muted,
                                modify_right=humanize_types):
            return False
    if value is not None:
        if not _basic_assertion(label, "Object is not equal to value", student_variable, value,
                                lambda l, r: equality_test(l, r, False, DELTA),
                                "{} != {}",
                                "was",
                                "to be equal to",
                                message, contextualize, report=report, partial_score=score, muted=muted,
                                show_expected_value=False):
            return False
    if report is None:
        report = MAIN_REPORT
    give_partial(score)
    return True


def assertGenerally(expression, score=None, message=None, report=MAIN_REPORT, muted=False,
                    contextualize=True, label="assert_generally"):
    """

    Args:
        expression:
        score:
        message:
        report:
        contextualize:

    Returns:

    """
    _setup_assertions(report)
    if expression:
        give_partial(score)
        return True
    else:
        report['assertions']['failures'] += 1
        if report['assertions']['exceptions']:
            raise AssertionException("General assertion")
        else:
            return False


# TODO: assertHasClass

# TODO: assertGraphType, assertGraphValues
assert_equal = assertEqual
assert_not_equal = assertNotEqual
assert_prints = assertPrints



from pedal.cait.cait_api import parse_program
from pedal.core.commands import explain, feedback
from pedal.core.feedback import AtomicFeedbackFunction, CompositeFeedbackFunction, Feedback
from pedal.toolkit.utilities import ensure_literal


# TODO
@AtomicFeedbackFunction(title="Opened Without Arguments",
                        message_template=("You have called the `open` function "
                                          "without any arguments. It needs a filename."))
def open_without_arguments(muted=False):
    """

    Args:
        muted:

    Returns:

    """
    return feedback(open_without_arguments.__name__, "toolkit", Feedback.CATEGORIES.INSTRUCTOR,
                    message=open_without_arguments.message_template.format(), muted=muted)


@CompositeFeedbackFunction()
def files_not_handled_correctly(*filenames, muted=False):
    """
    Statically detect if files have been opened and closed correctly.
    This is only useful in the case of very simplistic file handling.
    """
    if filenames and isinstance(filenames[0], int):
        num_filenames = filenames[0]
        actual_filenames = False
    else:
        num_filenames = len(filenames)
        actual_filenames = True
    ast = parse_program()
    calls = ast.find_all("Call")
    called_open = []
    closed = []
    for a_call in calls:
        if a_call.func.ast_name == 'Name':
            if a_call.func.id == 'open':
                if not a_call.args:
                    open_without_arguments(muted)
                    return True
                called_open.append(a_call)
            elif a_call.func.id == 'close':
                explain("You have attempted to call <code>close</code> as a "
                        "function, but it is actually a method of the "
                        "file object.", label="used_close_as_function", title="Close Is a Method", priority='verifier')
                return True
        elif a_call.func.ast_name == 'Attribute':
            if a_call.func.attr == 'open':
                explain("You have attempted to call <code>open</code> as a "
                        "method, but it is actually a built-in function.", label="used_open_as_method",
                        title="Open Is a Function")
                return True
            elif a_call.func.attr == 'close':
                closed.append(a_call)
    if len(called_open) < num_filenames:
        explain("You have not opened all the files you were supposed to.", label="unopened_files", title="Unopened Files")
        return True
    elif len(called_open) > num_filenames:
        explain("You have opened more files than you were supposed to.", label="extra_open_files", title="Extra Opened Files")
        return True
    withs = ast.find_all("With")
    if len(withs) + len(closed) < num_filenames:
        explain("You have not closed all the files you were supposed to.", label="unclosed_files", title="Unclosed Files")
        return True
    elif len(withs) + len(closed) > num_filenames:
        explain("You have closed more files than you were supposed to.", label="extra_closed_files", title="Extra Closed Files")
        return True
    if actual_filenames:
        return ensure_literal(*filenames)
    return False



import ast
import re
import sys
from pedal.core.report import MAIN_REPORT
from pedal.source.constants import TOOL_NAME
from pedal.source.feedbacks import not_enough_sections, incorrect_number_of_sections
from pedal.source.substitutions import Substitution

DEFAULT_SECTION_PATTERN = r'^(##### Part .+)$'


def separate_into_sections(pattern=DEFAULT_SECTION_PATTERN, report=MAIN_REPORT):
    """
    Chunks the current submissions' main code into separate sections.
    Args:
        pattern:
        report:

    Returns:

    """
    if not report[TOOL_NAME]['success']:
        pass
    if report[TOOL_NAME]['sections']:
        # TODO: System constraint violated: separating into sections multiple times
        pass
    report.group = 0
    report[TOOL_NAME]['section'] = 0
    report[TOOL_NAME]['line_offset'] = 0
    report[TOOL_NAME]['section_pattern'] = pattern
    report[TOOL_NAME]['sections'] = re.split(pattern, report.submission.main_code, flags=re.MULTILINE)

    backup = Substitution(report.submission.main_code, report.submission.main_file)
    report[TOOL_NAME]['substitutions'].append(backup)
    report.submission.replace_main(report[TOOL_NAME]['sections'][0])

    print(report[TOOL_NAME]['sections'])


def _calculate_section_number(section_index):
    return int((section_index+1)/2)


def stop_sections(report=MAIN_REPORT):
    """

    Args:
        report:
    """
    # TODO: If no substitutions, then throw error that we haven't separated into sections
    old_submission = report[TOOL_NAME]['substitutions'].pop()
    report.submission.replace_main(old_submission.code, old_submission.filename)


def next_section(name="", report=MAIN_REPORT):
    """

    Args:
        name:
        report:
    """
    report.execute_hooks(TOOL_NAME, 'next_section.before')
    source = report[TOOL_NAME]
    old_submission = report[TOOL_NAME]['substitutions'][-1]
    report.submission.replace_main(old_submission.code, old_submission.filename)
    # Advance to next section
    source['section'] += 2
    section_index = source['section']
    section_number = _calculate_section_number(section_index)
    sections = source['sections']
    found = len(source['sections'])
    if section_index < found:
        if source['independent']:
            new_code = ''.join(sections[section_index])
            old_code = ''.join(sections[:section_index])
            source['line_offset'] = len(old_code.split("\n"))-1
        else:
            new_code = ''.join(sections[:section_index + 1])
        report.submission.replace_main(new_code)
        report[TOOL_NAME]['success'] = None
        report.group = section_index
    else:
        not_enough_sections(section_number, found)
    report.execute_hooks(TOOL_NAME, 'next_section.after')


def check_section_exists(section_number, report=MAIN_REPORT):
    """
    Checks that the right number of sections exist. The prologue before the
    first section is 0, while subsequent ones are 1, 2, 3, etc. 
    So if you have 3 sections in your code plus the prologue,
    you should pass in 3 and not 4 to verify that all of them exist.
    """
    if report[TOOL_NAME]['success'] is False:
        return False
    found = int((len(report[TOOL_NAME]['sections']) - 1) / 2)
    print(section_number, found, len(report[TOOL_NAME]['sections']))
    if section_number > found:
        incorrect_number_of_sections(section_number, found, group=report[TOOL_NAME]['section'], report=report)


class _finish_section:
    def __init__(self, number, *functions):
        if isinstance(number, int):
            self.number = number
        else:
            self.number = -1
            functions = [number] + list(functions)
        self.functions = functions
        for function in functions:
            self(function, False)

    def __call__(self, f=None, quiet=True):
        if f is not None:
            f()
        if quiet:
            print("\tNEXT SECTION")

    def __enter__(self):
        pass

    def __exit__(self, x, y, z):
        print("\tNEXT SECTION")
        # return wrapped_f


def finish_section(number, *functions, **kwargs):
    """

    Args:
        number:
        *functions:
        **kwargs:

    Returns:

    """
    if 'next_section' in kwargs:
        ns = kwargs['next_section']
    else:
        ns = False
    if len(functions) == 0:
        x = _finish_section(number, *functions)
        x()
    else:
        result = _finish_section(number, *functions)
        if ns:
            print("\tNEXT SECTION")
        return result


# TODO: set up precondition and postcondition decorators for sections
def section(number):
    """
    """
    pass


def precondition(function):
    """

    Args:
        function:
    """
    pass


def postcondition(function):
    """

    Args:
        function:
    """
    pass



from pedal.environments.vpl_unittest import UnitTestedAssignment

"""
Some kind of function to break up the sections
"""
import re
import sys
from html.parser import HTMLParser

from pedal.core.report import MAIN_REPORT
from pedal import source
from pedal.resolvers import sectional
from pedal.cait.cait_api import expire_cait_cache


class VPLStyler(HTMLParser):
    """

    """
    HEADERS = ("h1", "h2", "h3", "h4", "h5")

    def __init__(self):
        super().__init__()
        self.reset()
        self.fed = []
        self.inside_pre = False

    def convert(self, html):
        """

        Args:
            html:

        Returns:

        """
        self.feed(html)
        return self.get_data()

    @property
    def text(self):
        """

        Returns:

        """
        return ''.join(self.fed)

    def get_data(self):
        """

        Returns:

        """
        return self.text

    def force_new_line(self):
        """

        """
        if self.text and self.text[-1] not in ("\n", "\r"):
            self.fed.append("\n")

    def handle_starttag(self, tag, attrs):
        """

        Args:
            tag:
            attrs:
        """
        if tag in self.HEADERS:
            self.force_new_line()
            self.fed.append("-")
        elif tag in ("pre",):
            self.force_new_line()
            self.fed.append(">")
            self.inside_pre = True

    def handle_data(self, data):
        """

        Args:
            data:
        """
        if self.inside_pre:
            # Need to prepend ">" to the start of new lines.
            self.fed.append(data.replace("\n", "\n>"))
        else:
            self.fed.append(data)

    def handle_endtag(self, tag):
        """

        Args:
            tag:
        """
        if tag in self.HEADERS:
            self.fed.append("")
        elif tag in ("pre",):
            self.fed.append("")
            self.inside_pre = False


def strip_tags(html):
    """

    Args:
        html:

    Returns:

    """
    return VPLStyler().convert(html)


def set_maximum_score(number, cap=True, report=None):
    """

    Args:
        number:
        cap:
        report:
    """
    if report is None:
        report = MAIN_REPORT
    report['vpl']['score_maximum'] = number
    report['vpl']['score_cap'] = cap


def resolve(report=None, custom_success_message=None):
    """

    Args:
        report:
        custom_success_message:
    """
    if report is None:
        report = MAIN_REPORT
    print("<|--")
    success, score, hc, messages_by_group = sectional.resolve(report)
    last_group = 0
    for group, messages in sorted(messages_by_group.items()):
        if group != last_group:
            for intermediate_section in range(last_group, group, 2):
                print("-" + report['source']['sections'][1 + intermediate_section])
        printed_first_bad = False
        for message in messages:
            if message['priority'] in ('positive', 'instructions'):
                print(strip_tags(message['message']))
            elif not printed_first_bad:
                print(strip_tags(message['message']))
                printed_first_bad = True
        last_group = group
    print("-Overall")
    if success:
        if custom_success_message is None:
            print("Complete! Great job!")
        else:
            print(custom_success_message)
    else:
        print("Incomplete")
    print("--|>")
    print("Grade :=>>", round(score))


class SectionalAssignment:
    """

    """
    max_points = 1
    sections = None

    def __init__(self, filename=None, max_points=None, report=None):
        self.report = MAIN_REPORT if report is None else report
        find_file(filename if filename else self.filename,
                  sections=True, report=report)
        set_maximum_score(self.max_points
                          if max_points is None else max_points)
        source.check_section_exists(self.sections)

    def pre_test(self):
        """

        Returns:

        """
        source.next_section()
        verified = source.verify_section()
        expire_cait_cache()
        return verified

    def post_test(self):
        """

        Returns:

        """
        return True

    def resolve(self):
        """

        """
        checks = ((self.pre_test() and
                   getattr(self, attr)() and
                   self.post_test())
                  for attr in dir(self)
                  if attr.startswith('test_') and
                  callable(getattr(self, attr)))
        if all(checks):
            self.report.set_success()
        resolve(report=self.report)


from pedal.environments.vpl_unittest import UnitTestedAssignment


def unittest_resolver(phases, report=None, custom_success_message=None):
    """

    Args:
        phases:
        report:
        custom_success_message:
    """
    success = True
    for title, phase in phases:
        outcome = phase()._run_all_tests()
        if not outcome:
            break
        success = success and outcome
    resolve(custom_success_message=custom_success_message)



import ast
import re

from pedal.cait.cait_api import parse_program
from pedal.cait.cait_node import CaitNode
from pedal.core.commands import gently, explain

"""
Verify indentation

Format:


Any number of text. One final newline separates the next section.

If line is "Args:" or "Returns:"
    Next line will be a "param (type): Description" or "type: Description"
    If the next line is indented more than current level, then it is part of the previous part's description.
    Otherwise, new entry

"Note:"
    Any level of indentation indicates
"""

PRIMITIVES = {
    'text': ['text'],
    'str': ['string', 'str', 'unicode'],
    'bytes': ['bytes'],
    'io': ['io'],
    'file': ['file'],
    'num': ['number', 'num', 'numeric'],
    'int': ['int', 'integer'],
    'float': ['float', 'floating'],
    'bool': ['bool', 'boolean'],
    'none': ['none'],
    'any': ['any']
}
NORMALIZE_PRIMITIVES = {synonym: formal
                        for formal, synonyms in PRIMITIVES.items()
                        for synonym in synonyms}
CONTAINERS = {
    'list': (1, ['list']),
    'set': (1, ['set']),
    'optional': (1, ['optional', 'maybe']),
    'dict': (2, ['dict', 'dictionary']),
    'callable': (2, ['callable', 'function', 'func']),
    'union': ('*', ['union', 'itemization']),
    'tuple': ('*', ['tuple', 'pair']),
}
NORMALIZE_CONTAINERS = {synonym: formal
                        for formal, (length, synonyms) in CONTAINERS.items()
                        for synonym in synonyms}

INHERITANCE = {
    'int': 'num',
    'float': 'num',
    'bool': 'num',
    'str': 'text',
    'bytes': 'text',
    'list': 'iterable',
    'tuple': 'iterable',
    'set': 'iterable',
    'dict': 'iterable',
    'file': 'iterable',
    'text': 'iterable'
}

SPECIAL_PARAMETERS = ["_returns", "yields", "prints", "_raises",
                      "_report", "_root"]

'''
Type validation:

* Caps does not matter
* Primitives:
* Containers
* Unions:
  * X or Y
  * X, Y, or Z
  * X, Y, Z
* Function: (X -> Y)
* list[int, str, or bool], dict[int: str], or bool or int
'''


def parse_type_slice(slice):
    """

    Args:
        slice:

    Returns:

    """
    if slice.ast_name == "Index":
        return parse_type(slice.value)
    elif slice.ast_name == "Slice":
        return "{}:{}".format(parse_type(slice.lower), parse_type(slice.upper))
    elif slice.ast_name == "ExtSlice":
        return ", ".join(parse_type_slice(s) for s in slice.dims)


def parse_type(node):
    """

    Args:
        node:

    Returns:

    """
    if node is None:
        return "Any"
    if node.ast_name == "Str":
        try:
            return parse_type(ast.parse(node.s).body[0].value)
        except:
            return node.s
    elif node.ast_name == "Name":
        return node.id
    elif node.ast_name == "NameConstant":
        return node.value
    elif node.ast_name == "List":
        return "[{}]".format(", ".join([parse_type(n) for n in node.elts]))
    elif node.ast_name == "Dict":
        return "{"+(", ".join(["{}: {}".format(parse_type(k), parse_type(v))
                                            for k,v in zip(node.keys, node.values)]))+"}"
    elif node.ast_name == "Subscript":
        return parse_type(node.value) + "[{}]".format(parse_type_slice(node.slice))
    elif node.ast_name == "BoolOp":
        if node.op.ast_name == "Or":
            return " or ".join(parse_type(v) for v in node.values)
    return "?"


def parse_type_value(value, parse_strings=False):
    """

    Args:
        value:
        parse_strings:

    Returns:

    """
    if isinstance(value, str):
        if parse_strings:
            return parse_type(CaitNode(ast.parse(value).body[0].value))
        else:
            return repr(value)
    elif value in (int, str, bool, float, list, dict, object):
        return value.__name__
    elif value is None:
        return "None"
    elif isinstance(value, list):
        if value:
            return "[{}]".format(parse_type_value(value[0]))
        else:
            return "[]"
    elif isinstance(value, tuple):
        if value == ():
            return "()"
        else:
            return "({})".format("".join(["{}, ".format(parse_type_value(v))
                                     for v in value]))
    elif isinstance(value, dict):
        if value == {}:
            return "{}"
        else:
            return "{" + (", ".join(["{}: {}".format(parse_type_value(k), parse_type_value(v))
                                     for k, v in value.items()])) + "}"


def test_type_equality(left, right):
    return left == right


class SignatureException(Exception):
    pass


class Stack:
    """

    """
    def __init__(self, identifier="union"):
        self.body = []
        self.identifier = identifier

    def append(self, value):
        """

        Args:
            value:
        """
        self.body.append(value)

    def __repr__(self):
        return "{}[{}]".format(self.identifier, ", ".join(map(repr, self.body)))

    def __hash__(self):
        return hash(tuple(self.identifier, self.body))

    def __lt__(self, other):
        if isinstance(other, Stack):
            return self.identifier < other.identifier and self.body < other.body
        return self.identifier < other

    def __gt__(self, other):
        if isinstance(other, Stack):
            return self.identifier > other.identifier and self.body > other.body
        return self.identifier > other

    def __eq__(self, other):
        if isinstance(other, Stack):
            return self.identifier == other.identifier and self.body == other.body
        return False


def _normalize_identifier(identifier):
    if identifier in NORMALIZE_PRIMITIVES:
        return NORMALIZE_PRIMITIVES[identifier]
    elif identifier in NORMALIZE_CONTAINERS:
        return NORMALIZE_CONTAINERS[identifier]
    else:
        return identifier


SPECIAL_SYMBOLS = r"\s*(->|\s*[\[\],\(\)\:\{\}]|or)\s*"


def _parse_tokens(tokens):
    result_stack = [Stack()]
    tokens = list(reversed(list(tokens)))
    while tokens:
        current = tokens.pop()
        # Ending a parenthetical, better stop here.
        if current == ")":
            subexpression = result_stack.pop()
            result_stack[-1].append(subexpression)
        # Ending a square bracket, better stop here.
        elif current == "]":
            subexpression = result_stack.pop()
            result_stack[-1].append(subexpression)
        # Ending a curly bracket, better stop here.
        elif current == "}":
            subexpression = result_stack.pop()
            result_stack[-1].append(subexpression)
        # We've reached the last token!
        elif not tokens:
            # And had no tokens before this one
            # Return the set of tokens
            result_stack[-1].append(_normalize_identifier(current))
        # Starting a parentheized expression
        elif current == "(":
            result_stack.append(Stack())
        elif current == "[":
            result_stack.append(Stack("list"))
        elif current == "{":
            result_stack.append(Stack("dict"))
        # Nullary function
        elif current == "->":
            result_stack[-1].append(Stack("callable"))
        elif current in ("or", ",", ":"):
            pass
        else:
            next = tokens.pop()
            # X or ...
            if current == "," and next == "or":
                tokens.append(next)
            if next in ("or", ",", "->", ":"):
                result_stack[-1].append(_normalize_identifier(current))
            # X [ ...
            elif next == "[":
                result_stack.append(Stack(_normalize_identifier(current)))
            else:
                tokens.append(next)
                result_stack[-1].append(_normalize_identifier(current))
    return result_stack.pop()


def sort_stacks(s):
    """

    Args:
        s:

    Returns:

    """
    if isinstance(s, Stack):
        return (True, (s.identifier, s.body))
    return (False, s)


def normalize_type(t):
    """

    Args:
        t:

    Returns:

    """
    t = t.strip()
    tokens = re.split(SPECIAL_SYMBOLS, t)
    tokens = [token for token in tokens if token]
    parsed = _parse_tokens(tokens)
    return parsed


def check_piece(left, right, indent=1):
    """

    Args:
        left:
        right:
        indent:

    Returns:

    """
    if type(left) != type(right):
        return False
    elif isinstance(left, Stack):
        if left.identifier != right.identifier:
            return False
        elif len(left.body) != len(right.body):
            return False
        elif left.identifier == "union":
            # Handle them in any order
            left.body.sort(key=sort_stacks)
            right.body.sort(key=sort_stacks)
        # Match them in exact order
        for l, r in zip(left.body, right.body):
            if not check_piece(l, r, indent=indent + 1):
                return False
        return True
    else:
        return left == right


def type_check(left, right):
    """

    Args:
        left:
        right:

    Returns:

    """
    left = normalize_type(left)
    right = normalize_type(right)
    return check_piece(left, right)
    
def find_colon(str):
    """

    Args:
        str:

    Returns:

    """
    parens_stack = []
    for i, character in enumerate(str):
        if character in '[(':
            parens_stack.append(character)
        elif character in '])':
            parens_stack.pop()
        elif character == ':' and not parens_stack:
            return i
    return 0
    
ARGS = ('args:', 'arg:', 'argument:', 'arguments:',
        'parameters:', 'params:', 'parameter:', 'param:')
ARG_PATTERN = r'(.+)\s*\((.+)\)\s*:(.+)'
RETURNS = ('returns:', 'return:')
def parse_docstring(doc):
    """

    Args:
        doc:

    Returns:

    """
    # First line's indentation may be different from rest - trust first
    # non empty line after the first one.
    # Remove taht number of spaces from subsequent lines
    # If Line is "Args:" or other special...
    # 
    lines = doc.split("\n")
    body = [lines[0]]
    args = {}
    current_arg = None
    returns = []
    current_component = 'body'
    indentation = None
    inner_indentation = None
    for line in lines[1:]:
        # Blank line, not interesting!
        if not line.strip():
            continue
        # Get the actual text
        if indentation is None:
            indentation = len(line) - len(line.lstrip())
        line = line[indentation:]
        potential_command = line.lower().strip()
        # New command region?
        if potential_command in ARGS:
            current_component = 'args'
            inner_indentation = None
            continue
        elif potential_command in RETURNS:
            current_component = 'returns'
            inner_indentation = None
            continue
        # Okay, it's content - let's process it
        if current_component == 'body':
            body.append(line)
        else:
            if inner_indentation is None:
                inner_indentation = len(line) - len(line.lstrip())
            line = line[inner_indentation:]
            # Skip indented lines
            if not re.match(r'\s', line):
                if current_component == 'args':
                    match = re.search(ARG_PATTERN, line)
                    current_arg = match.group(1)
                    type_str = match.group(2)
                    args[current_arg.strip()] = type_str.strip()
                elif current_component == 'returns':
                    position = find_colon(line)
                    return_type, comment = line[:position], line[position:]
                    returns.append(return_type.strip())
    return body, args, ' or '.join(returns)


def function_signature(function_name, returns=None, yields=None,
                       prints=None, raises=None, report=None, root=None,
                       **kwargs):
    """
    Determines whether the function with this signature is in the AST.
    
    TODO: Implement raises, prints, yields
    """
    if root is None:
        root = parse_program()
    # If you encounter any special parameters with a "_", then fix their
    # name. This allows for students to have parameters with the given name.
    for special_parameter in SPECIAL_PARAMETERS:
        if special_parameter in kwargs:
            kwargs[special_parameter[1:]] = kwargs.pop(special_parameter)
    # Go get the actual docstring, parse it
    docstring = None
    for function_def in root.find_all("FunctionDef"):
        if function_def._name == function_name:
            if function_def.body:
                if (function_def.body[0].ast_name == "Expr" and
                        function_def.body[0].value.ast_name == "Str"):
                    docstring = function_def.body[0].value.s
    # Try to match each element in turn.
    if docstring is None:
        return False

    try:
        body, args, parsed_returns = parse_docstring(docstring)
    except Exception as e:
        return [e], False
    failing_parameters = []
    for name, type in kwargs.items():
        if name in args:
            if not type_check(type, args[name]):
                failing_parameters.append(name)
        else:
            failing_parameters.append(name)
    if returns is None and not returns:
        return failing_parameters, True
    elif returns is not None and returns:
        return failing_parameters, type_check(parsed_returns, returns)
    else:
        return failing_parameters, False
    

def class_signature(class_name, report=None, root=None, **attributes):
    """

    Args:
        class_name:
        **attributes:
        report:
        root:

    Returns:

    """
    if root is None:
        root = parse_program()


"""

"""



__all__ = ['Resolver']


class Resolver:
    def resolve(self):
        """

        """
        pass

    def _on_pre_resolve(self):
        """
        Hook event.
        Returns:

        """

    def _on_post_resolve(self):
        """
        Hook event.
        Returns:

        """



from pedal.core.commands import feedback
from pedal.core.feedback import AtomicFeedbackFunction
from pedal.core.report import MAIN_REPORT
from pedal.utilities.exceptions import ExpandedTraceback
from pedal.source.constants import TOOL_NAME


@AtomicFeedbackFunction(title="No Source Code",
                        text_template="Source code file is blank.",
                        justification="After stripping the code, there were no characters.")
def blank_source(group=None, report=MAIN_REPORT):
    """

    Args:
        group:
        report:

    Returns:

    """
    return feedback(blank_source.__name__, category=feedback.CATEGORIES.SYNTAX, tool=TOOL_NAME,
                    title=blank_source.title, version=blank_source.version,
                    group=group, message=blank_source.text_template(), report=report)


@AtomicFeedbackFunction(title="Not Enough Sections",
                        text_template=("Tried to advance to next section but the "
                                       "section was not found. Tried to load section "
                                       "{count}, but there were only {found} sections."),
                        justification="Section index exceeded the length of the separated sections list.")
def not_enough_sections(section_number, found, report=MAIN_REPORT):
    """

    Args:
        section_number:
        found:
        report:

    Returns:

    """
    fields = {'count': section_number, 'found': found}
    return feedback(not_enough_sections.__name__, category=feedback.CATEGORIES.SYNTAX, tool=TOOL_NAME,
                    title=not_enough_sections.title, version=not_enough_sections.version,
                    fields=fields,
                    message=not_enough_sections.text_template(**fields), report=report)


@AtomicFeedbackFunction(title='Source File Not Found',
                        text_template=("The given filename ('{filename}') was either not found"
                                       " or could not be opened. Please make sure the file is"
                                       " available."),
                        version='0.0.1',
                        justification="IOError while opening file to set_source")
def source_file_not_found(filename, sections, report=MAIN_REPORT):
    """

    Args:
        filename:
        sections:
        report:

    Returns:

    """
    fields = {'filename': filename, 'sections': sections}
    return feedback(source_file_not_found.__name__, category=feedback.CATEGORIES.SYNTAX, tool=TOOL_NAME,
                    title=source_file_not_found.title, version=source_file_not_found.version,
                    fields=fields,
                    text=source_file_not_found.text_template(**fields),
                    group=0 if sections else None,
                    report=report)


@AtomicFeedbackFunction(title="Syntax Error",
                        message_template="```\n{context}\n```\nBad syntax on line {lineno}",
                        text_template="{context}\nBad syntax on line {lineno}",
                        version='0.0.1',
                        justification="Syntax error was triggered while calling ast.parse")
def syntax_error(line, filename, code, offset, text, traceback, exception, exc_info, report=MAIN_REPORT, muted=False):
    """

    Args:
        line:
        filename:
        code:
        offset:
        text:
        traceback:
        exception:
        exc_info:
        report:
        muted:

    Returns:

    """
    traceback = ExpandedTraceback(exception, exc_info, traceback, report.submission.instructor_file,
                                  line, filename, code)
    context = traceback.format_exception()
    fields = {'lineno': line, 'filename': filename, 'offset': offset, 'text': text,
              'traceback': traceback, 'exception': exception, 'context': context}
    return feedback(syntax_error.__name__, category=feedback.CATEGORIES.SYNTAX, tool=TOOL_NAME,
                    title=syntax_error.title, version=syntax_error.version,
                    fields=fields,
                    message=syntax_error.message_template(**fields),
                    text=syntax_error.text_template(**fields), muted=muted,
                    locations=line,
                    report=report)


@AtomicFeedbackFunction(title="Incorrect Number of Sections",
                        text_template=("Incorrect number of sections in your file. "
                                       "Expected {count}, but only found {found}"),
                        justification="")
def incorrect_number_of_sections(count, found, group, report=MAIN_REPORT):
    """

    Args:
        count:
        found:
        group:
        report:

    Returns:

    """
    fields = {'count': count, 'found': found}
    return feedback(incorrect_number_of_sections.__name__, category=feedback.CATEGORIES.SYNTAX, tool=TOOL_NAME,
                    title=incorrect_number_of_sections.title, fields=fields, group=group,
                    text=incorrect_number_of_sections.text_template(**fields), report=report)


# TODO: IndentationError
# TODO: TabError



"""
A simple environment for quickly running Pedal code.

Parameter Order
    Files
    MainFile
    MainCode
    User
    Assignment
    Course
    Execution
    InstructorFile

Get submission (Main Code, Main Filename, Other Code, Other Settings) from
    as argument
        main_code, main_file
        files = {'answer.py': code}

    sys.args
        filename
        parameter
    global variable
    filename

Provide Output
    STDOUT
    Global variable
    Filename
    Returned result


"""

import sys
from pedal.core.commands import contextualize_report
from pedal.core.submission import Submission
from pedal.source import verify
from pedal.cait import parse_program
from pedal.sandbox import run
from pedal.tifa import tifa_analysis
from pedal.resolvers import simple


def parse_argv():
    if len(sys.argv) == 2:
        main_file = sys.argv[1]
        with open(main_file) as main_file_handle:
            main_code = main_file_handle.read()
        return sys.argv[0], {main_file: main_code}, main_file, main_code, None, None, None, None
    #elif len(sys.argv) == 3:
    else:
        return sys.argv
    # TODO: Finish handling arguments intelligently


def setup_pedal(files=None, main_file='answer.py', main_code=None,
                user=None, assignment=None, course=None, execution=None,
                instructor_file='on_run.py', skip_tifa=False):
    if files is None and main_code is None:
        instructor_file, files, main_file, main_code, user, assignment, course, execution = parse_argv()
    elif files is None:
        files = {main_file: main_code}
    elif main_code is None:
        main_code = files[main_file]
    contextualize_report(Submission(files, main_file, main_code, user, assignment, course, execution,
                                    instructor_file))
    verify()
    ast = parse_program()
    if not skip_tifa:
        tifa = tifa_analysis()
    student = run(threaded=True, raise_exceptions=True)
    student.report_exceptions_mode = True

    def print_resolve(*args, **kwargs):
        result = simple.resolve(*args, **kwargs)
        print(result.label)
        print(result.score)
        print(result.title)
        print(result.message)
        return result

    return ast, student, print_resolve



import traceback
import os

BuiltinKeyError = KeyError


class KeyError(BuiltinKeyError):
    """
    A version of KeyError that replaces the built-in with one small
    modification: when printing an explanatory message, the message is not
    rendered as a tuple. Because that's stupid and the fact that it made it
    into CPython is just rude.

    See Also:
        https://github.com/python/cpython/blob/master/Objects/exceptions.c#L1556
    """
    __module__ = "builtins"

    def __init__(self, original, message):
        for field in ['__cause__', '__traceback__', '__context__']:
            if hasattr(original, field):
                setattr(self, field, getattr(original, field))
            else:
                setattr(self, field, None)
        self.message = message

    def __str__(self):
        return self.message


def add_context_to_error(e, message):
    """

    Args:
        e:
        message:

    Returns:

    """
    if isinstance(e, BuiltinKeyError):
        new_args = repr(e.args[0]) + message
        e = KeyError(e, new_args)
        e.args = tuple([new_args])
    elif isinstance(e, OSError):
        # TODO: Investigate OSError, since they have so many args.
        #       Might be weird.
        e.args = tuple([e.args[0] + message])
        return e
    elif hasattr(e, 'args') and e.args:
        e.args = tuple([e.args[0] + message])
    return e


class ExpandedTraceback:
    """
    Class for reformatting tracebacks to have more pertinent information.
    """

    def __init__(self, exception, exc_info, full_traceback,
                 instructor_filename, line_offset, student_filename,
                 original_code_lines):
        """
        Args:
            exception (Exception): The exception that was raised.
            exc_info (ExcInfo): The result of sys.exc_info() when the exception
                was raised.
            full_traceback (bool): Whether or not to provide the full traceback
                or just the parts relevant to students.
            instructor_filename (str): The name of the instructor file, which
                can be used to avoid reporting instructor code in the
                traceback.
        """
        self.line_offset = line_offset
        self.exception = exception
        self.exc_info = exc_info
        self.full_traceback = full_traceback
        self.instructor_filename = instructor_filename
        self.student_filename = student_filename
        self.line_number = traceback.extract_tb(exc_info[2])[-1][1]
        self.original_code_lines = original_code_lines

    @staticmethod
    def _clean_traceback_line(line):
        return line.replace(', in <module>', '', 1)

    def format_exception(self, preamble=""):
        """

        Args:
            preamble:

        Returns:

        """
        if not self.exception:
            return ""
        if isinstance(self.exception, TimeoutError):
            return str(self.exception)
        cl, exc, tb = self.exc_info
        while tb and self._is_relevant_tb_level(tb):
            tb = tb.tb_next
        length = self._count_relevant_tb_levels(tb)
        tb_e = traceback.TracebackException(cl, self.exception, tb, limit=length,
                                            capture_locals=False)
        # print(list(), file=x)
        for frame in tb_e.stack:
            if frame.filename == os.path.basename(self.student_filename):
                frame.lineno += self.line_offset
            if frame.lineno - 1 < len(self.original_code_lines):
                frame._line = self.original_code_lines[frame.lineno - 1]
            else:
                frame._line = "*line missing*"
        lines = [self._clean_traceback_line(line)
                 for line in tb_e.format()]
        lines[0] = "Traceback:\n"
        return preamble + ''.join(lines)

    def _count_relevant_tb_levels(self, tb):
        length = 0
        while tb and not self._is_relevant_tb_level(tb):
            length += 1
            tb = tb.tb_next
        return length

    def _is_relevant_tb_level(self, tb):
        """
        Determines if the give part of the traceback is relevant to the user.

        Returns:
            boolean: True means it is NOT relevant
        """
        # Are in verbose mode?
        if self.full_traceback:
            return False
        filename, a_, b_, _ = traceback.extract_tb(tb, limit=1)[0]
        # Is the error in the student file?
        if filename == self.student_filename:
            return False
        # Is the error in the instructor file?
        if filename == self.instructor_filename:
            return True
        # Is the error in this test directory?
        current_directory = os.path.dirname(os.path.realpath(__file__))
        if filename.startswith(current_directory):
            return True
        # Is the error related to a file in the parent directory?
        # parent_directory = os.path.dirname(current_directory)
        # TODO: Currently we don't refer to this?
        # Is the error in a local file?
        if filename.startswith('.'):
            return False
        # Is the error in an absolute path?
        if not os.path.isabs(filename):
            return False
        # Okay, it's not a student related file
        return True



import ast

from pedal.tifa.feedbacks import append_to_non_list


def are_literals_equal(first, second):
    """

    Args:
        first:
        second:

    Returns:

    """
    if first is None or second is None:
        return False
    elif not isinstance(first, type(second)):
        return False
    else:
        if isinstance(first, LiteralTuple):
            if len(first.value) != len(second.value):
                return False
            for l, s in zip(first.value, second.value):
                if not are_literals_equal(l, s):
                    return False
            return True
        elif not isinstance(first, LiteralValue):
            return True
        else:
            return first.value == second.value


class LiteralValue:
    """
    A special literal representation of a value, used to represent access on
    certain container types.
    """

    def __init__(self, value):
        self.value = value


class LiteralNum(LiteralValue):
    """
    Used to capture indexes of containers.
    """

    def type(self):
        """

        Returns:

        """
        return NumType()


class LiteralBool(LiteralValue):
    def type(self):
        """

        Returns:

        """
        return BoolType()


class LiteralStr(LiteralValue):
    def type(self):
        """

        Returns:

        """
        return StrType()


class LiteralTuple(LiteralValue):
    def type(self):
        """

        Returns:

        """
        return TupleType(self.value)


class LiteralNone(LiteralValue):
    def type(self):
        """

        Returns:

        """
        return LiteralNone()


def literal_from_json(val):
    """

    Args:
        val:

    Returns:

    """
    if val['type'] == 'LiteralStr':
        return LiteralStr(val['value'])
    elif val['type'] == 'LiteralNum':
        return LiteralNum(val['value'])
    elif val['type'] == 'LiteralBool':
        return LiteralBool(val['value'])


def _dict_extends(d1, d2):
    """
    Helper function to create a new dictionary with the contents of the two
    given dictionaries. Does not modify either dictionary, and the values are
    copied shallowly. If there are repeates, the second dictionary wins ties.

    The function is written to ensure Skulpt compatibility.

    Args:
        d1 (dict): The first dictionary
        d2 (dict): The second dictionary
    """
    d3 = {}
    for key, value in d1.items():
        d3[key] = value
    for key, value in d2.items():
        d3[key] = value
    return d3


class Type:
    """
    Parent class for all other types, used to provide a common interface.

    TODO: Handle more complicated object-oriented types and custom types
    (classes).
    """
    fields = {}
    immutable = False
    singular_name = 'a type'

    def clone(self):
        """

        Returns:

        """
        return self.__class__()

    def __str__(self):
        return str(self.__class__.__name__)

    def precise_description(self):
        """

        Returns:

        """
        return self.singular_name

    def clone_mutably(self):
        """

        Returns:

        """
        if self.immutable:
            return self.clone()
        else:
            return self

    def index(self, i):
        """

        Args:
            i:

        Returns:

        """
        return self.clone()

    def iterate(self, i):
        """

        Args:
            i:

        Returns:

        """
        return self.index(i)

    def load_attr(self, attr, tifa, callee=None, callee_position=None):
        """

        Args:
            attr:
            tifa:
            callee:
            callee_position:

        Returns:

        """
        if attr in self.fields:
            return self.fields[attr]
        # TODO: Handle more kinds of common mistakes
        if attr == "append":
            tifa._issue(append_to_non_list(callee_position, tifa.identify_caller(callee), self, report=tifa.report))
        return UnknownType()

    def is_empty(self):
        """

        Returns:

        """
        return True

    def is_equal(self, other):
        """

        Args:
            other:

        Returns:

        """
        # TODO: Make this more sophisticated!
        if type(self) not in TYPE_LOOKUPS:
            return False
        return other in TYPE_LOOKUPS[type(self)]

    def is_instance(self, other):
        """

        Args:
            other:

        Returns:

        """
        # TODO: Implement this correctly
        return self.is_equal(other)


class UnknownType(Type):
    """
    A special type used to indicate an unknowable type.
    """


class RecursedType(Type):
    """
    A special type used as a placeholder for the result of a
    recursive call that we have already process. This type will
    be dominated by any actual types, but will not cause an issue.
    """


class FunctionType(Type):
    """
    
    Special values for `returns`:
        identity: Returns the first argument's type
        element: Returns the first argument's first element's type
        void: Returns the NoneType
    """
    singular_name = 'a function'

    def __init__(self, definition=None, name="*Anonymous", returns=None):
        if returns is not None and definition is None:
            if returns == 'identity':
                def definition(ti, ty, na, args, ca):
                    """

                    Args:
                        ti:
                        ty:
                        na:
                        args:
                        ca:

                    Returns:

                    """
                    if args:
                        return args[0].clone()
                    return UnknownType()
            elif returns == 'element':
                def definition(ti, ty, na, args, ca):
                    """

                    Args:
                        ti:
                        ty:
                        na:
                        args:
                        ca:

                    Returns:

                    """
                    if args:
                        return args[0].index(0)
                    return UnknownType()
            elif returns == 'void':
                def definition(ti, ty, na, args, ca):
                    """

                    Args:
                        ti:
                        ty:
                        na:
                        args:
                        ca:

                    Returns:

                    """
                    return NoneType()
            else:
                def definition(ti, ty, na, args, ca):
                    """

                    Args:
                        ti:
                        ty:
                        na:
                        args:
                        ca:

                    Returns:

                    """
                    return returns.clone()
        self.definition = definition
        self.name = name


class ClassType(Type):
    """

    """
    singular_name = 'a class'

    def __init__(self, name):
        self.name = name
        self.fields = {}
        self.scope_id = None

    def add_attr(self, name, type):
        """

        Args:
            name:
            type:
        """
        self.fields[name] = type

    def get_constructor(self):
        """

        Returns:

        """
        i = InstanceType(self)
        return FunctionType(name='__init__', returns=i)

    def clone(self):
        """

        Returns:

        """
        return ClassType(self.name)


class InstanceType(Type):
    """

    """
    def __init__(self, parent):
        self.parent = parent
        self.fields = parent.fields

    def __str__(self):
        return "InstanceTypeOf" + str(self.parent.name)

    def clone(self):
        """

        Returns:

        """
        return InstanceType(self.parent)

    def add_attr(self, name, type):
        """

        Args:
            name:
            type:
        """
        # TODO: What if this is a type change?
        self.parent.add_attr(name, type)


class NumType(Type):
    singular_name = 'a number'
    immutable = True

    def index(self, i):
        """

        Args:
            i:

        Returns:

        """
        return UnknownType()


class NoneType(Type):
    singular_name = 'a None'
    immutable = True


class BoolType(Type):
    singular_name = 'a boolean'
    immutable = True


class TupleType(Type):
    """
    """
    singular_name = 'a tuple'

    def __init__(self, subtypes=None):
        if subtypes is None:
            subtypes = []
        self.subtypes = subtypes

    def index(self, i):
        """

        Args:
            i:

        Returns:

        """
        if isinstance(i, LiteralNum):
            return self.subtypes[i.value].clone()
        else:
            return self.subtypes[i].clone()

    def clone(self):
        """

        Returns:

        """
        return TupleType([t.clone() for t in self.subtypes])

    immutable = True


class ListType(Type):
    """

    """
    singular_name = 'a list'

    def __init__(self, subtype=None, empty=True):
        if subtype is None:
            subtype = UnknownType()
        self.subtype = subtype
        self.empty = empty

    def index(self, i):
        """

        Args:
            i:

        Returns:

        """
        return self.subtype.clone()

    def clone(self):
        """

        Returns:

        """
        return ListType(self.subtype.clone(), self.empty)

    def load_attr(self, attr, tifa, callee=None, callee_position=None):
        """

        Args:
            attr:
            tifa:
            callee:
            callee_position:

        Returns:

        """
        if attr == 'append':
            def _append(tifa, function_type, callee, args, position):
                if args:
                    cloned_type = ListType(subtype=args[0].clone(),
                                           empty=False)
                    if callee:
                        tifa.append_variable(callee, cloned_type, position)
                    self.empty = False
                    self.subtype = args[0]

            return FunctionType(_append, 'append')
        return Type.load_attr(self, attr, tifa, callee, callee_position)

    def is_empty(self):
        """

        Returns:

        """
        return self.empty


class StrType(Type):
    """

    """
    singular_name = 'a string'

    def __init__(self, empty=False):
        self.empty = empty

    def index(self, i):
        """

        Args:
            i:

        Returns:

        """
        return StrType()

    def is_empty(self):
        """

        Returns:

        """
        return self.empty

    fields = _dict_extends(Type.fields, {})
    immutable = True


StrType.fields.update({
    # Methods that return strings
    "capitalize": FunctionType(name='capitalize', returns=StrType()),
    "center": FunctionType(name='center', returns=StrType()),
    "expandtabs": FunctionType(name='expandtabs', returns=StrType()),
    "join": FunctionType(name='join', returns=StrType()),
    "ljust": FunctionType(name='ljust', returns=StrType()),
    "lower": FunctionType(name='lower', returns=StrType()),
    "lstrip": FunctionType(name='lstrip', returns=StrType()),
    "replace": FunctionType(name='replace', returns=StrType()),
    "rjust": FunctionType(name='rjust', returns=StrType()),
    "rstrip": FunctionType(name='rstrip', returns=StrType()),
    "strip": FunctionType(name='strip', returns=StrType()),
    "swapcase": FunctionType(name='swapcase', returns=StrType()),
    "title": FunctionType(name='title', returns=StrType()),
    "translate": FunctionType(name='translate', returns=StrType()),
    "upper": FunctionType(name='upper', returns=StrType()),
    "zfill": FunctionType(name='zfill', returns=StrType()),
    # Methods that return numbers
    "count": FunctionType(name='count', returns=NumType()),
    "find": FunctionType(name='find', returns=NumType()),
    "rfind": FunctionType(name='rfind', returns=NumType()),
    "index": FunctionType(name='index', returns=NumType()),
    "rindex": FunctionType(name='rindex', returns=NumType()),
    # Methods that return booleans
    "startswith": FunctionType(name='startswith', returns=BoolType()),
    "endswith": FunctionType(name='endswith', returns=BoolType()),
    "isalnum": FunctionType(name='isalnum', returns=BoolType()),
    "isalpha": FunctionType(name='isalpha', returns=BoolType()),
    "isdigit": FunctionType(name='isdigit', returns=BoolType()),
    "islower": FunctionType(name='islower', returns=BoolType()),
    "isspace": FunctionType(name='isspace', returns=BoolType()),
    "istitle": FunctionType(name='istitle', returns=BoolType()),
    "isupper": FunctionType(name='isupper', returns=BoolType()),
    # Methods that return List of Strings
    "rsplit": FunctionType(name='rsplit', returns=ListType(StrType(), empty=False)),
    "split": FunctionType(name='split', returns=ListType(StrType(), empty=False)),
    "splitlines": FunctionType(name='splitlines', returns=ListType(StrType(), empty=False))
})


class FileType(Type):
    singular_name = 'a file'

    def index(self, i):
        """

        Args:
            i:

        Returns:

        """
        return StrType()

    fields = _dict_extends(Type.fields, {
        'close': FunctionType(name='close', returns='void'),
        'read': FunctionType(name='read', returns=StrType()),
        'readlines': FunctionType(name='readlines', returns=ListType(StrType(), False))
    })

    def is_empty(self):
        """

        Returns:

        """
        return False


class DictType(Type):
    """

    """
    singular_name = 'a dictionary'

    def precise_description(self):
        """

        Returns:

        """
        base = "a dictionary"
        if self.literals:
            base += " mapping "
            # TODO: Handle recursive precise names more correctly
            base += ", ".join("{!r} to {}".format(l.value, r.precise_description())
                              for l, r in zip(self.literals, self.values))
        elif self.keys:
            keys = self.keys[0] if isinstance(self.keys, list) else self.keys
            values = self.values[0] if isinstance(self.values, list) else self.values
            base += " mapping {}".format(keys.precise_description())
            base += " to {}".format(values.precise_description())
        return base

    def __init__(self, empty=False, literals=None, keys=None, values=None):
        self.empty = empty
        self.literals = literals
        self.values = values
        self.keys = keys

    def clone(self):
        """

        Returns:

        """
        return DictType(self.empty, self.literals, self.keys, self.values)

    def is_empty(self):
        """

        Returns:

        """
        return self.empty

    def has_literal(self, l):
        """

        Args:
            l:

        Returns:

        """
        for literal, value in zip(self.literals, self.values):
            if are_literals_equal(literal, l):
                return value
        return None

    def _access_item(self, i, get_keys=False):
        if self.empty:
            return UnknownType()
        elif self.literals is not None:
            for literal, value in zip(self.literals, self.values):
                if are_literals_equal(literal, i):
                    return value.clone()
            return UnknownType()
        elif get_keys:
            # TODO: Handle multiple keys/values
            return self.keys[0].clone()
        else:
            return self.values[0].clone()

    def index(self, i):
        """

        Args:
            i:

        Returns:

        """
        return self._access_item(i, False)

    def iterate(self, i):
        """

        Args:
            i:

        Returns:

        """
        return self._access_item(i, True)

    def update_key(self, literal_key, type):
        """

        Args:
            literal_key:
            type:
        """
        self.literals.append(literal_key)
        self.values.append(type)


    def load_attr(self, attr, tifa, callee=None, callee_position=None):
        """

        Args:
            attr:
            tifa:
            callee:
            callee_position:

        Returns:

        """
        if attr == 'items':
            def _items(tifa, function_type, callee, args, position):
                if self.literals is None:
                    return ListType(TupleType([self.keys, self.values]),
                                    empty=False)
                else:
                    return ListType(TupleType([self.literals[0].type(),
                                               self.values[0]]),
                                    empty=False)

            return FunctionType(_items, 'items')
        elif attr == 'keys':
            def _keys(tifa, function_type, callee, args, position):
                if self.literals is None:
                    return ListType(self.keys, empty=False)
                else:
                    return ListType(self.literals[0].type(), empty=False)

            return FunctionType(_keys, 'keys')
        elif attr == 'values':
            def _values(tifa, function_type, callee, args, position):
                if self.literals is None:
                    return ListType(self.values, empty=False)
                else:
                    return ListType(self.values[0], empty=False)

            return FunctionType(_values, 'values')
        return Type.load_attr(self, attr, tifa, callee, callee_position)


class ModuleType(Type):
    """

    """
    singular_name = 'a module'

    def __init__(self, name="*UnknownModule", submodules=None, fields=None):
        self.name = name
        if submodules is None:
            submodules = {}
        self.submodules = submodules
        if fields is None:
            fields = {}
        self.fields = fields


class SetType(ListType):
    singular_name = 'a set'


class GeneratorType(ListType):
    singular_name = 'a generator'


# Custom parking class in blockpy

class TimeType(Type):
    singular_name = 'a time of day'


class DayType(Type):
    singular_name = 'a day of the week'


try:
    from numbers import Number
except Exception:
    Number = int

TYPE_LOOKUPS = {
    FunctionType: ('function', FunctionType, 'FunctionType'),
    ClassType: ('class', ClassType, 'ClassType'),
    InstanceType: ('instance', InstanceType, 'InstanceType'),
    NumType: ('num', int, float, complex, NumType, Number, 'NumType'),
    BoolType: ('bool', bool, BoolType, 'BoolType'),
    NoneType: ('None', None, NoneType, 'NoneType'),
    TupleType: ('tuple', tuple, TupleType, 'TupleType'),
    ListType: ('list', list, ListType, 'ListType'),
    StrType: ('str', str, StrType, 'StrType'),
    FileType: ('file', FileType, 'FileType'),
    DictType: ('dict', dict, DictType, 'DictType'),
    SetType: ('set', set, SetType, 'SetType'),
}


def type_from_json(val):
    """

    Args:
        val:

    Returns:

    """
    if val['type'] == 'DictType':
        values = [type_from_json(v) for v in val['values']]
        empty = val.get('empty', None)
        if 'literals' in val:
            literals = [literal_from_json(l) for l in val['literals']]
            return DictType(empty, literals=literals, values=values)
        else:
            keys = [type_from_json(k) for k in val['keys']]
            return DictType(empty, keys=keys, values=values)
    elif val['type'] == 'ListType':
        return ListType(type_from_json(val.get('subtype', None)),
                        val.get('empty', None))
    elif val['type'] == 'StrType':
        return StrType(val.get('empty', None))
    elif val['type'] == 'BoolType':
        return BoolType()
    elif val['type'] == 'NoneType':
        return NoneType()
    elif val['type'] == 'NumType':
        return NumType()
    elif val['type'] == 'ModuleType':
        submodules = {name: type_from_json(m)
                      for name, m in val.get('submodules', {}).items()}
        fields = {name: type_from_json(m)
                  for name, m in val.get('fields', {}).items()}
        return ModuleType(name=val.get('name'), submodules=submodules,
                          fields=fields)
    elif val['type'] == 'FunctionType':
        returns = type_from_json(val.get('returns', {'type': 'NoneType'}))
        return FunctionType(name=val.get('name'), returns=returns)


def type_to_literal(type):
    """

    Args:
        type:

    Returns:

    """
    if isinstance(type, NumType):
        return LiteralNum(0)
    elif isinstance(type, StrType):
        return LiteralStr("")
    else:
        # TODO: Finish the mapping
        return LiteralStr("")


TYPE_STRINGS = {
    "str": StrType, "string": StrType,
    "num": NumType, "number": NumType, "int": NumType, "integer": NumType, "float": NumType,
    "complex": NumType,
    "bool": BoolType, "boolean": BoolType,
    "none": NoneType,
    "dict": DictType, "dictionary": DictType,
    "list": ListType,
    "tuple": TupleType,
    "set": SetType,
    "file": FileType,
    "func": FunctionType, "function": FunctionType,
    "class": ClassType,
}


def get_tifa_type_from_str(value, self):
    """

    Args:
        value:
        self:

    Returns:

    """
    #if value in custom_types:
    #    return custom_types[value]
    if value.lower() in TYPE_STRINGS:
        return TYPE_STRINGS[value.lower()]()
    else:
        variable = self.find_variable_scope(value)
        if variable.exists:
            state = self.load_variable(value)
            return state.type
        #custom_types.add(value)
        return UnknownType()
        # TODO: handle custom types


def get_tifa_type(v, self):
    """

    Args:
        v:
        self:

    Returns:

    """
    if isinstance(v, ast.Str):
        return get_tifa_type_from_str(v.s, self)
    elif isinstance(v, ast.Name):
        return get_tifa_type_from_str(v.id, self)
    elif isinstance(v, ast.List):
        elements = v.elts
        if elements:
            return ListType(subtype=get_tifa_type(elements[0], self))
        else:
            return ListType(empty=True)
    elif isinstance(v, ast.Dict):
        if not v.keys:
            return DictType(empty=True)
        if all(isinstance(k, ast.Str) for k in v.keys):
            return DictType(literals=[LiteralStr(s.s) for s in v.keys],
                            values=[get_tifa_type(vv, self) for vv in v.values])
        return DictType(keys=[get_tifa_type(k, self) for k in v.keys],
                        values=[get_tifa_type(vv, self) for vv in v.values])
    # TODO: Finish filling in static type system
    else:
        return UnknownType()



def humanize_type(t):
    """

    Args:
        t:

    Returns:

    """
    if hasattr(t, '__name__'):
        return t.__name__
    else:
        return str(t)


def humanize_types(types):
    """

    Args:
        types:

    Returns:

    """
    if isinstance(types, tuple):
        return ', '.join(humanize_type(t) for t in types)
    return humanize_type(types)


from pedal.toolkit.utilities import function_is_called
from pedal.cait.cait_api import parse_program, def_use_error
from pedal.core.commands import gently, explain
from pedal.sandbox import compatibility

PLOT_LABEL = {'plot': 'line plot',
              'hist': 'histogram',
              'scatter': 'scatter plot'}


def prevent_incorrect_plt():
    """

    Returns:

    """
    ast = parse_program()
    plts = [n for n in ast.find_all("Name") if n.id == 'plt']
    if plts and def_use_error(plts[0]):
        # TODO: I converted this to the explain_r function, but I wasn't sure about the priority thing ~Luke Gusukuma
        # explain("You have imported the <code>matplotlib.pyplot</code> module, "
        #         "but you did not rename it to <code>plt</code> using "
        #         "<code>import matplotlib.pyplot as plt</code>.<br><br><i>(plt_rename_err)<i></br></br>", 'verifier')
        explain("You have imported the <code>matplotlib.pyplot</code> module, "
                  "but you did not rename it to <code>plt</code> using "
                  "<code>import matplotlib.pyplot as plt</code>.",
                  label="plt_rename_err",
                  priority='syntax')
        return True
    matplotlib_names = ['plot', 'hist', 'scatter',
                        'title', 'xlabel', 'ylabel', 'show']
    for name in matplotlib_names:
        for n in ast.find_all("Name"):
            if n.id == name:
                if def_use_error(n):
                    # explain(("You have attempted to use the MatPlotLib "
                    #          "function named <code>{0}</code>. However, you "
                    #          "imported MatPlotLib in a way that does not "
                    #          "allow you to use the function directly. I "
                    #          "recommend you use <code>plt.{0}</code> instead, "
                    #          "after you use <code>import matplotlib.pyplot as "
                    #          "plt</code>.<br><br><i>(plt_wrong_import)<i></br></br>").format(name), 'verifier')
                    explain(("You have attempted to use the MatPlotLib "
                               "function named <code>{0}</code>. However, you "
                               "imported MatPlotLib in a way that does not "
                               "allow you to use the function directly. I "
                               "recommend you use <code>plt.{0}</code> instead, "
                               "after you use <code>import matplotlib.pyplot as "
                               "plt</code>.").format(name),
                              label="plt_wrong_import",
                              priority='syntax')
                    return True
    return False


def ensure_correct_plot(function_name):
    """

    Args:
        function_name:

    Returns:

    """
    for a_plot, label in PLOT_LABEL.items():
        if function_name == a_plot:
            if not function_is_called(function_name):
                gently("You are not calling the <code>{func_name}</code> function.".format(func_name=function_name),
                       title="Not calling {func_name}".format(func_name=function_name),
                       fields={'func_name': function_name},
                       label="no_{func_name}_call".format(func_name=function_name))
                return True
        elif function_is_called(a_plot):
            gently("You have called the <code>{}</code> function, which makes a {}.".format(a_plot, label),
                     label="wrong_plt")
            return True
    return False


def ensure_show():
    """

    Returns:

    """
    if not function_is_called("show"):
        gently("You have not called <code>show</code> function, which "
                 "actually creates the graph.", label="no_show")
        return True
    return False


def compare_data(plt_type, correct, given):
    """
    Determines whether the given data matches any of the data found in the
    correct data. This handles plots of different types: if a histogram
    was plotted with the expected data for a line plot, it will return True.

    Args:
        plt_type (str): The expected type of this plot
        correct (List of Int or List of List of Int): The expected data.
        given (Dict): The actual plotted data and information
    Returns:
        bool: Whether the correct data was found in the given plot.
    """
    # Infer arguments
    if plt_type == 'hist':
        correct_xs = None
        correct_ys = correct
    elif not correct:
        correct_xs = []
        correct_ys = []
    elif isinstance(correct[0], (tuple, list)):
        # We were given a list of lists of ints
        correct_xs, correct_ys = correct
    else:
        # Assume it is a singular list
        correct_xs = list(range(len(correct)))
        correct_ys = correct

    if given['type'] == 'hist':
        return correct_ys == given['values']
    elif plt_type == 'hist':
        return correct_ys == given['y']
    else:
        return correct_xs == given['x'] and correct_ys == given['y']


GRAPH_TYPES = {'line': 'line plot',
               'hist': 'histogram',
               'scatter': 'scatter plot'}


def check_for_plot(plt_type, data, muted=True):
    """
    TODO: Convert to be feedback function!

    Returns any errors found for this plot type and data.
    In other words, if it returns False, the plot was found correctly.
    """
    if plt_type == 'plot':
        plt_type = 'line'
    type_found = False
    data_found = False
    for graph in compatibility.get_plots():
        for a_plot in graph['data']:
            data_found_here = compare_data(plt_type, data, a_plot)
            if a_plot['type'] == plt_type and data_found_here:
                return False
            if a_plot['type'] == plt_type:
                type_found = True
            if data_found_here:
                data_found = True
    plt_type = GRAPH_TYPES.get(plt_type, plt_type)
    if type_found and data_found:
        return ("You have created a {}, but it does not have the right data. That data appears to have been plotted "
                "in another graph.<br><br><i>(other_plt)<i></br></br>".format(plt_type))
    elif type_found:
        return ("You have created a {}, but it does not have the right data."
                "<br><br><i>(wrong_plt_data)<i></br></br>".format(plt_type))
    elif data_found:
        return ("You have plotted the right data, but you appear to have not plotted it as a {}."
                "<br><br><i>(wrong_plt_type)<i></br></br>".format(plt_type))
    else:
        return ("You have not created a {} with the proper data."
                "<br><br><i>(no_plt)<i></br></br>".format(plt_type))


def check_for_plot_r(plt_type, data):
    """
    Returns any errors found for this plot type and data.
    In other words, if it returns False, the plot was found correctly.
    """
    if plt_type == 'plot':
        plt_type = 'line'
    type_found = False
    data_found = False
    for graph in compatibility.get_plots():
        for a_plot in graph['data']:
            data_found_here = compare_data(plt_type, data, a_plot)
            if a_plot['type'] == plt_type and data_found_here:
                return False
            if a_plot['type'] == plt_type:
                type_found = True
            if data_found_here:
                data_found = True
    plt_type = GRAPH_TYPES.get(plt_type, plt_type)
    if type_found and data_found:
        return {"message": "You have created a {}, but it does not have the right data. "
                           "That data appears to have been plotted in another graph.".format(plt_type),
                "code": "other_plt",
                "label": "Plotting Another Graph"}
    elif type_found:
        return {"message": "You have created a {}, but it does not have the right data.".format(plt_type),
                "code": "wrong_plt_data",
                "label": "Plot Data Incorrect"}
    elif data_found:
        return {"message": "You have plotted the right data, but you appear to have not plotted it as a {}.".format(plt_type),
                "code": "wrong_plt_type",
                "label": "Wrong Plot Type"
                }
    else:
        return {"message": "You have not created a {} with the proper data.".format(plt_type),
                "code": "no_plt",
                "label": "Missing Plot"}



from pedal.core.commands import feedback
from pedal.core.report import MAIN_REPORT
from pedal.core.feedback import Feedback, CompositeFeedbackFunction
from pedal.sandbox import TOOL_NAME

@CompositeFeedbackFunction()
def runtime_error(message, name, exception, position, text=None, muted=False, group=None, report=MAIN_REPORT):
    """

    Args:
        message:
        name:
        exception:
        position:
        text:
        muted:
        group:
        report:

    Returns:

    """
    fields = {'message': message,
              'name': name,
              'error': exception,
              'position': position,
              'traceback': None}
    # TODO: Finish
    return feedback(name, tool=TOOL_NAME, category=Feedback.CATEGORIES.RUNTIME, kind=Feedback.KINDS.MISTAKE,
                    justification="A runtime error occurred during execution of some code", title=name,
                    message=message, text=text or message, fields=fields, locations=position, muted=muted,
                    version='0.0.1', group=group, report=report)
